/*
  regex.c - Regex pattern matching library

  Copyright (C) 2001, 2002 Ville Laurikari <vl@iki.fi>.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License version 2 (June
  1991) as published by the Free Software Foundation.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

/*
  The algorithm used in this library uses linear worst-case time in
  the length of the text being searched, and quadratic worst-case time
  in the length of the used regular expression.  In other words, the
  complexity of the algorithm is O(|e|^2 * |w|), where `e' is the
  regular expression and `w' is the text.  The used space is also
  quadratic on the length of the regex, but does not depend on the
  searched string.  This quadratic behaviour occurs only on
  pathological cases which are probably very rare in practice.

  It would be possible to implement an algorithm with better space and
  time complexity, but I presume the overhead of handling a functional
  map of subexpression start and end positions would make the matcher
  just slower in practice.  Computing epsilon closures instead of
  using epsilon-free automata would in theory take asymptotically less
  time and memory, but with practical patterns it would just slow
  things down.

  Currently back referencing is not supported.  The problem of finding
  matches for regular expressions with back referencing is NP-complete
  (on the number of back references) so most likely an efficient generic
  algorithm does not exist, nor will I spend any time on looking for
  one.  Instead, I will just write a simple backtracking thing when I get
  the time.

  This code has not been divided into several files mostly because GNU
  regex, and some other implementations, only have one file.  This way
  it is easier to replace those other implementations with libtre by just
  dropping this in place of the old `regex.c'.

  Warning: currently this code is at places documented very poorly,
  and what little there is may be outdated, confusing, or even
  misleading.

*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif /* HAVE_CONFIG_H */

#include <assert.h>
#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stddef.h>
#include <limits.h>
#include <errno.h>
#ifdef HAVE_WCHAR_H
#include <wchar.h>
#endif /* HAVE_WCHAR_H */
#ifdef HAVE_WCTYPE_H
#include <wctype.h>
#endif /* HAVE_WCTYPE_H */     
#ifdef HAVE_LOCALE_H
#include <locale.h>
#endif /* HAVE_LOCALE_H */

#ifdef DEBUG
#include <stdio.h>
#define DPRINT(msg) do { } while(0)
//#define DPRINT(msg) do {printf msg; fflush(stdout);} while(0)
#else /* !DEBUG */
#define DPRINT(msg) do { } while(0)
#endif /* !DEBUG */

#include "memory.h" // changed by Luis Figueiredo (i've changed malloc,free,realloc,calloc to mymalloc,myfree,myrealloc,mycalloc)
#include "debug.h"
#include "regex.h"
//#undef DEBUG

/*
  Memory management.  This is for allocating small memory blocks efficiently
  in terms of memory overhead and execution speed.  The allocated blocks
  cannot be freed individually, only all at once.
*/

typedef struct tre_list {
  void *data;
  struct tre_list *next;
} tre_list_t;

typedef struct {
  tre_list_t *blocks;
  tre_list_t *current;
  char *ptr;
  size_t n;
  int failed;
} tre_mem_t;

#define TRE_MEM_BLOCK_SIZE 1024

static tre_mem_t *
tre_mem_new(void)
{
  tre_mem_t *mem;

  mem = mycalloc(1, sizeof(*mem));
  if (mem == NULL)
    return NULL;
  return mem;
}

/* Allocates a block of `size' bytes from `mem'.  Returns a pointer to the
   allocated block or NULL if an underlying malloc() failed or `size' is
   larger than TRE_MEM_BLOCK_SIZE.  */
static void *
tre_mem_alloc(tre_mem_t *mem, size_t size)
{
  void *ptr;
  size += size % sizeof(long);  /* For alignment. */
  if (size > TRE_MEM_BLOCK_SIZE)
    {
      DPRINT(("tre_mem_alloc: size exceeds maximum\n"));
      return NULL;
    }
  if (mem->failed)
    {
      DPRINT(("tre_mem_alloc: oops, called after failure?!\n"));
      return NULL;
    }

#ifdef MALLOC_DEBUGGING

  ptr = mymalloc(1);
  if (ptr == NULL)
    {
      DPRINT(("tre_mem_alloc: malloc forced failure\n"));
      mem->failed = 1;
      return NULL;
    }

  myfree(ptr);
#endif /* MALLOC_DEBUGGING */

  if (mem->n < size)
    {
      /* Allocate a new block. */
      tre_list_t *l;
      DPRINT(("tre_mem_alloc: allocating new block\n"));

      l = mymalloc(sizeof(*l));
      if (l == NULL)
	{
	  mem->failed = 1;
	  return NULL;
	}

      l->data = mymalloc(TRE_MEM_BLOCK_SIZE);
      if (l->data == NULL)
	{

	  myfree(l);
	  mem->failed = 1;
	  return NULL;
	}
      l->next = NULL;
      if (mem->current != NULL)
	mem->current->next = l;
      if (mem->blocks == NULL)
	mem->blocks = l;
      mem->current = l;
      mem->ptr = l->data;
      mem->n = TRE_MEM_BLOCK_SIZE;
    }

  /* Allocate from current block. */
  ptr = mem->ptr;
  mem->ptr += size;
  mem->n -= size;
  return ptr;
}


static void
tre_mem_free(tre_mem_t *mem)
{
  tre_list_t *tmp, *l = mem->blocks;

  while (l != NULL)
    {

      myfree(l->data);
      tmp = l->next;

      myfree(l);
      l = tmp;
    }

  myfree(mem);
}



/*
  Stack functions.
*/

typedef struct {
  int size;
  int max_size;
  int increment;
  int ptr;
  void **stack;
} tre_stack_t;

static tre_stack_t *
stack_new(int size, int max_size, int increment)
{
  tre_stack_t *s;


  s = mymalloc(sizeof(*s));
  if (s != NULL)
    {

      s->stack = mymalloc(sizeof(*s->stack) * size);
      if (s->stack == NULL)
	{

	  myfree(s);
	  return NULL;
	}
      s->size = size;
      s->max_size = max_size;
      s->increment = increment;
      s->ptr = 0;
    }
  return s;
}

static void
stack_destroy(tre_stack_t *s)
{

  myfree(s->stack);

  myfree(s);
}

static int
stack_size(tre_stack_t *s)
{
  return s->ptr;
}

static reg_errcode_t
stack_push(tre_stack_t *s, void *value)
{
  if (s->ptr < s->size)
    {
      s->stack[s->ptr] = value;
      s->ptr++;
    }
  else
    {
      if (s->size >= s->max_size)
	{
	  DPRINT(("stack_push: stack full\n"));
	  return REG_ESPACE;
	}
      else
	{
	  void **new_buffer;
	  int new_size;
	  DPRINT(("stack_push: trying to realloc more space\n"));
	  new_size = s->size + s->increment;
	  if (new_size > s->max_size)
	    new_size = s->max_size;

	  new_buffer = myrealloc(s->stack, sizeof(*new_buffer) * new_size);
	  if (new_buffer == NULL)
	    {
	      DPRINT(("stack_push: realloc failed.\n"));
	      return REG_ESPACE;
	    }
	  DPRINT(("stack_push: realloc succeeded.\n"));
	  assert(new_size > s->size);
	  s->size = new_size;
	  s->stack = new_buffer;
	  stack_push(s, value);
	}
    }
  return REG_OK;
}

/* Just to save some typing. */
#define STACK_PUSH(s, value)                                                  \
  do									      \
    {									      \
      status = stack_push(s, (void *)(value));				      \
    }                                                                         \
  while (0);

#define STACK_PUSHX(s, value)                                                 \
  {									      \
    status = stack_push(s, (void *)(value));				      \
    if (status != REG_OK)                                                     \
      break;                                                                  \
  }


static void *
stack_pop(tre_stack_t *s)
{
  return s->stack[--s->ptr];
}



/*
  Regex AST stuff.
*/

typedef enum { LITERAL, CATENATION, ITERATION, UNION } tre_ast_type_t;

/* Special leaf types. */
#define EMPTY     -1   /* Empty leaf (denotes empty string). */
#define ASSERTION -2   /* Assertion leaf. */
#define TAG       -3   /* Tag leaf. */

/* Assertions. */
#define ASSERT_AT_BOL             1          /* Beginning of line. */
#define ASSERT_AT_EOL             2          /* End of line. */
#define ASSERT_CHAR_CLASS         4          /* Character class. */
#define ASSERT_CHAR_CLASS_NEG     8          /* Character classes, negated. */
#define ASSERT_AT_BOW            16          /* Beginning of word. */
#define ASSERT_AT_EOW            32          /* End of word. */
#define ASSERT_AT_WB             64          /* Word boundary. */
#define ASSERT_AT_WB_NEG        128          /* Word boundary, negated. */
#define ASSERT_LAST             128

//typedef unsigned long int wctype_t; // By Luis Figueiredo TAGL

typedef struct {
  int position;
  int code_min;
  int code_max;
  int *tags;
  int assertions;
  wctype_t class;
  wctype_t *neg_classes;
} tre_pos_and_tags_t;

typedef struct {
  tre_ast_type_t type;
  void *obj;
  int nullable;
  int submatch_id;
  int num_submatches;
  int num_tags;
  tre_pos_and_tags_t *firstpos;
  tre_pos_and_tags_t *lastpos;
} tre_ast_node_t;

typedef struct {
  long code_min;
  long code_max;
  int position;
  wctype_t class;
  wctype_t *neg_classes;
} tre_literal_t;

typedef struct {
  tre_ast_node_t *left;
  tre_ast_node_t *right;
} tre_catenation_t;

typedef struct {
  tre_ast_node_t *arg;
  int min;
  int max;
} tre_iteration_t;

typedef struct {
  tre_ast_node_t *left;
  tre_ast_node_t *right;
} tre_union_t;

static tre_ast_node_t *
ast_new_node(tre_mem_t *mem, tre_ast_type_t type, size_t size)
{
  tre_ast_node_t *node;

  node = tre_mem_alloc(mem, sizeof(*node));
  if (node == NULL)
    return NULL;
  node->obj = tre_mem_alloc(mem, size);
  if (node->obj == NULL)
    return NULL;
  node->type = type;
  node->nullable = -1;
  node->submatch_id = -1;
  node->num_submatches = 0;
  node->num_tags = 0;
  node->firstpos = NULL;
  node->lastpos = NULL;

  return node;
}

static tre_ast_node_t *
ast_new_literal(tre_mem_t *mem, int code_min, int code_max, int position)
{
  tre_ast_node_t *node;

  node = ast_new_node(mem, LITERAL, sizeof(tre_literal_t));
  if (node == NULL)
    return NULL;
  ((tre_literal_t *)node->obj)->code_min = code_min;
  ((tre_literal_t *)node->obj)->code_max = code_max;
  ((tre_literal_t *)node->obj)->position = position;
  ((tre_literal_t *)node->obj)->class = (wctype_t)0;
  ((tre_literal_t *)node->obj)->neg_classes = NULL;

  return node;
}

static tre_ast_node_t *
ast_new_iteration(tre_mem_t *mem, tre_ast_node_t *arg, int min, int max)
{
  tre_ast_node_t *node;

  node = ast_new_node(mem, ITERATION, sizeof(tre_iteration_t));
  if (node == NULL)
    return NULL;
  ((tre_iteration_t *)node->obj)->arg = arg;
  ((tre_iteration_t *)node->obj)->min = min;
  ((tre_iteration_t *)node->obj)->max = max;
  node->num_submatches = arg->num_submatches;

  return node;
}

static tre_ast_node_t *
ast_new_union(tre_mem_t *mem, tre_ast_node_t *left, tre_ast_node_t *right)
{
  tre_ast_node_t *node;

  node = ast_new_node(mem, UNION, sizeof(tre_union_t));
  if (node == NULL)
    return NULL;
  ((tre_union_t *)node->obj)->left = left;
  ((tre_union_t *)node->obj)->right = right;
  node->num_submatches = left->num_submatches + right->num_submatches;

  return node;
}

static tre_ast_node_t *
ast_new_catenation(tre_mem_t *mem, tre_ast_node_t *left, tre_ast_node_t *right)
{
  tre_ast_node_t *node;

  node = ast_new_node(mem, CATENATION, sizeof(tre_catenation_t));
  if (node == NULL)
    return NULL;
  ((tre_catenation_t *)node->obj)->left = left;
  ((tre_catenation_t *)node->obj)->right = right;
  node->num_submatches = left->num_submatches + right->num_submatches;

  return node;
}

#ifdef REGEX_DEBUG

static void
findent(FILE *stream, int i)
{
  while (i-- > 0)
    fputc(' ', stream);
}


static void
do_print(FILE *stream, tre_ast_node_t *ast, int indent)
{
  int code_min, code_max, pos;

  findent(stream, indent);
  switch (ast->type)
    {
    case LITERAL:
      code_min = ((tre_literal_t *)ast->obj)->code_min;
      code_max = ((tre_literal_t *)ast->obj)->code_max;
      pos = ((tre_literal_t *)ast->obj)->position;
      if (code_min == EMPTY)
	{
	  fprintf(stream, "literal empty\n");
	}
      else if (code_min == ASSERTION)
	{
	  int i;
	  char *assertions[] = { "bol", "eol", "ctype", "!ctype",
				 "bow", "eow", "wb", "!wb" };
	  if (code_max >= ASSERT_LAST << 1)
	    assert(0);
	  fprintf(stream, "assertions: ");
	  for (i = 0; (1 << i) <= ASSERT_LAST; i++)
	    if (code_max & (1 << i))
	      fprintf(stream, "%s ", assertions[i]);
	  fprintf(stream, "\n");
	}
      else if (code_min == TAG)
	{
	  fprintf(stream, "tag %d\n", code_max);
	}
      else
	{
	  fprintf(stream, "literal (%c, %c) (%d, %d), pos %d, sub %d\n",
		  code_min, code_max, code_min, code_max, pos,
		  ast->submatch_id);
	}
      break;
    case ITERATION:
      fprintf(stream, "iteration {%d, %d}, sub %d\n",
	      ((tre_iteration_t *)ast->obj)->min,
	      ((tre_iteration_t *)ast->obj)->max,
	      ast->submatch_id);
      do_print(stream, ((tre_iteration_t *)ast->obj)->arg, indent + 2);
      break;
    case UNION:
      fprintf(stream, "union, sub %d\n", ast->submatch_id);
      do_print(stream, ((tre_union_t *)ast->obj)->left, indent + 2);
      do_print(stream, ((tre_union_t *)ast->obj)->right, indent + 2);
      break;
    case CATENATION:
      fprintf(stream, "catenation, sub %d\n", ast->submatch_id);
      do_print(stream, ((tre_catenation_t *)ast->obj)->left, indent + 2);
      do_print(stream, ((tre_catenation_t *)ast->obj)->right, indent + 2);
      break;
    default:
      assert(0);
      break;
    }
}

static void
ast_fprint(FILE *stream, tre_ast_node_t *ast)
{
  do_print(stream, ast, 0);
}

static void
ast_print(tre_ast_node_t *tree)
{
  printf("AST:\n");
  ast_fprint(stdout, tree);
}

#endif /* REGEX_DEBUG */



/*
  Algorithms to setup tags so that submatch addressing can be done.
*/


/* Inserts a catenation node to the root of the tree given in `node'.
   As the left child a new tag with number `tag_id' to `node' is added,
   and the right child is the old root. */
static reg_errcode_t
add_tag_left(tre_mem_t *mem, tre_ast_node_t *node, int tag_id)
{
  tre_catenation_t *c;

  DPRINT(("add_tag_left: tag %d\n", tag_id));

  c = tre_mem_alloc(mem, sizeof(*c));
  if (c == NULL)
    return REG_ESPACE;
  c->left = ast_new_literal(mem, TAG, tag_id, -1);
  if (c->left == NULL)
    return REG_ESPACE;
  c->right = tre_mem_alloc(mem, sizeof(tre_ast_node_t));
  if (c->right == NULL)
    return REG_ESPACE;

  c->right->obj = node->obj;
  c->right->type = node->type;
  c->right->nullable = -1;
  c->right->submatch_id = -1;
  c->right->firstpos = NULL;
  c->right->lastpos = NULL;
  node->obj = c;
  node->type = CATENATION;

  return REG_OK;
}

/* Inserts a catenation node to the root of the tree given in `node'.
   As the right child a new tag with number `tag_id' to `node' is added,
   and the left child is the old root. */
static reg_errcode_t
add_tag_right(tre_mem_t *mem, tre_ast_node_t *node, int tag_id)
{
  tre_catenation_t *c;

  DPRINT(("add_tag_right: tag %d\n", tag_id));

  c = tre_mem_alloc(mem, sizeof(*c));
  if (c == NULL)
    return REG_ESPACE;
  c->right = ast_new_literal(mem, TAG, tag_id, -1);
  if (c->right == NULL)
    return REG_ESPACE;
  c->left = tre_mem_alloc(mem, sizeof(tre_ast_node_t));
  if (c->left == NULL)
    return REG_ESPACE;

  c->left->obj = node->obj;
  c->left->type = node->type;
  c->left->nullable = -1;
  c->left->submatch_id = -1;
  c->left->firstpos = NULL;
  c->left->lastpos = NULL;
  node->obj = c;
  node->type = CATENATION;

  return REG_OK;
}


/* Instructions to compute submatch register values after a successful
   match.  */
struct submatch_data {
  /* The start offset of this submatch can be computed by taking the tag
     specified by `so_tag' and subtracting `so_offset'. */
  int so_offset;
  int so_tag;
  /* The end offset of this submatch can be computed by taking the tag
     specified by `eo_tag' and subtracting `eo_offset'. */
  int eo_offset;
  int eo_tag;
  /* marker_tag > 0: If the value of the tag specified by `marker_tag' is -1
     then this subexpression does not participate in the match.
     marker_tag = 0: This subexpression always participates in the match.  */
  int marker_tag;
  /* `marker_boundary' > 0: If the value of the tag specified by `marker_tag'
     is smaller than the value of the tag specified by `marker_boundary'
     then this subexpression does not participate in the match, even if
     the marker tag is positive. */
  int marker_boundary;
};

typedef struct submatch_data tre_submatch_data_t;


static void
move_regs(int *regset, tre_submatch_data_t *submatch_data, int new_tag)
{
  int i;

  for (i = 0; regset[i] >= 0; i++)
    {
      DPRINT(("move_regs: moving submatch %d %s reg to t%d\n",
	      regset[i] / 2, regset[i] % 2 ? "end" : "start",
	      new_tag));
      if (regset[i] % 2)
	submatch_data[regset[i]/2].eo_tag = new_tag;
      else
	submatch_data[regset[i]/2].so_tag = new_tag;
    }
}

typedef enum {
  ADDTAGS_RECURSE,
  ADDTAGS_SET_SUBMATCH_END,
  ADDTAGS_CATENATION_NEXT,
  ADDTAGS_POST_CATENATION,
  ADDTAGS_POST_ITERATION,
  ADDTAGS_UNION_NEXT,
  ADDTAGS_POST_UNION
} tre_addtags_symbol_t;

#define TAG_MINIMIZE 0
#define TAG_MAXIMIZE 1

/* Adds tag leafs to appropriate locations to the syntax tree given in
   `node', so that subexpressions marked for submatch addressing can be
   traced.  */
static reg_errcode_t
ast_add_tags(tre_mem_t *mem, tre_stack_t *stack, tre_ast_node_t *node,
	     int groups, tre_submatch_data_t *submatch_data,
	     int *tag_directions, int *end_tag, int *tag_min_bounds,
	     int *marker_offs)
{
  reg_errcode_t status = REG_OK;
  tre_addtags_symbol_t symbol;
  int *regset, *orig_regset;
  int curr_tag = 0;            /* current tag not yet added */
  int curr_marker_tag = 0;     /* first tag in union operands */
  int moving_marker = 0;
  int curr_marker_bound = -1;  /* tag just before union */
  int next_marker_bound = -1;
  int curr_boundary_tag = -1;  /* tag just before iteration */
  int next_tag = 1;
  int bottom;
  int num_tags = node->num_tags;


  regset = mymalloc(sizeof(*regset) * ((groups + 1) * 2));
  if (regset == NULL)
    return REG_ESPACE;
  regset[0] = -1;
  orig_regset = regset;

  if (tag_directions != NULL)
    tag_directions[0] = TAG_MINIMIZE;
  bottom = stack_size(stack);
  STACK_PUSH(stack, node);
  STACK_PUSH(stack, ADDTAGS_RECURSE);

  while (stack_size(stack) > bottom)
    {
      if (status != REG_OK)
	break;

      symbol = (tre_addtags_symbol_t)stack_pop(stack);
      switch (symbol)
	{
	case ADDTAGS_SET_SUBMATCH_END:
	  {
	    int i, submatch_id = (int)stack_pop(stack);
	    IFDEBUG(fprintf(stderr,"here?"));
	    DPRINT(("ast_add_tags: ) using tag %d for submatch %d end\n",
		    curr_tag, submatch_id));
	    submatch_data[submatch_id].eo_offset = 0;
	    submatch_data[submatch_id].eo_tag = curr_tag;
	    for (i = 0; regset[i] >= 0; i++);
	    regset[i] = submatch_id * 2 + 1;
	    regset[i + 1] = -1;
	    break;
	  }

	case ADDTAGS_RECURSE:
	  node = stack_pop(stack);
#if 0
	  DPRINT(("\nast_add_tags: %p ", node));
	  ast_print(node);
#endif /* REGEX_DEBUG */
	  if (node->submatch_id >= 0)
	    {
	      int i, submatch_id = node->submatch_id;

	      DPRINT(("ast_add_tags: ( using tag %d for submatch %d start, "
		      "marker tag %d, marker boundary tag %d\n",
		      curr_tag, submatch_id, curr_marker_tag,
		      curr_marker_bound));
	      submatch_data[submatch_id].so_offset = 0;
	      submatch_data[submatch_id].so_tag = curr_tag;
	      submatch_data[submatch_id].marker_tag = curr_marker_tag;
	      submatch_data[submatch_id].marker_boundary = curr_marker_bound;
	      for (i = 0; regset[i] >= 0; i++);
	      regset[i] = submatch_id * 2;
	      regset[i + 1] = -1;

	      /* When rising back up from this subtree, add the ending
		 point to the tag similarly. */
	      STACK_PUSHX(stack, node->submatch_id);
	      STACK_PUSHX(stack, ADDTAGS_SET_SUBMATCH_END);
	    }
	  switch (node->type)
	    {
	    case LITERAL:
	      {
		tre_literal_t *lit = (tre_literal_t *)node->obj;
		int i;

		if (lit->code_min >= 0)
		  {
		    DPRINT(("ast_add_tags: literal %d-%d\n",
			    (int)lit->code_min, (int)lit->code_max));
		    /* Increase the offset of each register in the set. */
		    for (i = 0; regset[i] >= 0; i++)
		      {
			DPRINT(("ast_add_tags: increasing reg %d offset\n",
				regset[i]));
			if (regset[i] % 2)
			  submatch_data[regset[i] / 2].eo_offset++;
			else
			  submatch_data[regset[i] / 2].so_offset++;
		      }
		    /* Increase the offset of the marker tag too. */
		    if (curr_marker_tag > 0 && moving_marker)
		      {
			marker_offs[curr_marker_tag]++;
			DPRINT(("ast_add_tags: marker_offs[%d] = %d\n",
				curr_marker_tag,
				marker_offs[curr_marker_tag]));
		      }
		  }
		else
		  {
		    /* There can't be a tag in the tree here because we haven't
		       added any yet. */
		    assert(lit->code_min == ASSERTION
			   || lit->code_min == EMPTY);
		  }
		break;
	      }
	    case ITERATION:
	      {
		tre_ast_node_t *arg = ((tre_iteration_t *)node->obj)->arg;

		DPRINT(("ast_add_tags: iteration, current tag %d, "
			"next tag %d\n", curr_tag, next_tag));
		/* Tags cannot be pushed over iteration operators, so we
		   create a tag for the the current register set here. */
		status = add_tag_left(mem, node, curr_tag);
		if (status != REG_OK)
		  break;

		/* Store the current boundary tag value. */
		STACK_PUSHX(stack, curr_boundary_tag);
		STACK_PUSHX(stack, ADDTAGS_POST_ITERATION);
		curr_boundary_tag = curr_tag;

		if (arg->submatch_id >= 0)
		  {
		    DPRINT(("ast_add_tags: left boundary for pmatch[%d].so is "
			    "given by tag %d\n", arg->submatch_id,
			    curr_boundary_tag));
		    tag_min_bounds[arg->submatch_id * 2] = curr_boundary_tag;
		  }

		/* Start a new register set. */
		curr_tag = next_tag;
		regset[0] = -1;
		if (curr_tag < num_tags)
		  tag_directions[curr_tag] = TAG_MINIMIZE;
		next_tag++;

		/* Whatever this means...*/
		moving_marker = 0;
		next_marker_bound = curr_tag;

		STACK_PUSHX(stack, arg);
		STACK_PUSHX(stack, ADDTAGS_RECURSE);
		break;
	      }

	    case UNION:
	      {
		int have_submatches = (node->num_submatches
				       - (node->submatch_id >= 0 ? 1 : 0)) > 0;
		tre_ast_node_t *left = ((tre_union_t *)node->obj)->left;
		tre_ast_node_t *right = ((tre_union_t *)node->obj)->right;
		int orig_next_tag = next_tag;

		/* XXX - If both operands always match exactly the same
		   length strings we can push the current tag over
		   this node! */

		curr_marker_bound = next_marker_bound;
		DPRINT(("ast_add_tags: union, curr = %d, next %d, "
			"marker = %d, bound = %d\n", curr_tag, next_tag,
			curr_marker_tag, curr_marker_bound));

		if (regset[0] != -1 || have_submatches)
		  {
		    /* Tags cannot be pushed over union operators, so we
		       create a tag for the current tag set here. */
		    status = add_tag_left(mem, node, curr_tag);
		    if (status != REG_OK)
		      break;
		    regset[0] = -1;
		    curr_tag = next_tag;
		    next_tag++;
		  }

		/* If there are no subexpressions marked for submatch
		   addressing there is no need to recurse into this branch. */
		if (!have_submatches)
		  {
		    DPRINT(("ast_add_tags: no submatches inside, skipping\n"));
		    break;
		  }

		/* After processing right child. */
		STACK_PUSHX(stack, curr_marker_tag);
		STACK_PUSHX(stack, regset);
		STACK_PUSHX(stack, orig_next_tag + left->num_tags
			    + right->num_tags + (left->num_tags ? 0 : 1)
			    + (right->num_tags ? 0 : 1));
		STACK_PUSHX(stack, right);
		STACK_PUSHX(stack, right->num_tags);
		STACK_PUSHX(stack, ADDTAGS_POST_UNION);
		/* Process right child. */
		STACK_PUSHX(stack, right);
		STACK_PUSHX(stack, ADDTAGS_RECURSE);
		/* After processing left child. */
		STACK_PUSHX(stack, orig_next_tag + left->num_tags
			    + (left->num_tags ? 0 : 1));
		STACK_PUSHX(stack, left);
		STACK_PUSHX(stack, left->num_tags);
		STACK_PUSHX(stack, ADDTAGS_UNION_NEXT);
		/* Process left child. */
		STACK_PUSHX(stack, left);
		STACK_PUSHX(stack, ADDTAGS_RECURSE);

		curr_marker_tag = curr_tag;
		moving_marker = 1;
		regset[0] = -1;
		if (curr_tag < num_tags)
		  tag_directions[curr_tag] = TAG_MAXIMIZE;
		break;
	      }

	    case CATENATION:
	      {
		tre_ast_node_t *left = ((tre_catenation_t *)node->obj)->left;
		tre_ast_node_t *right = ((tre_catenation_t *)node->obj)->right;
		int reserved_tag = -1;
		int orig_next_tag = next_tag;

		DPRINT(("ast_add_tags: catenation start left, current tag %d, "
			"next tag %d\n", curr_tag, next_tag));
		DPRINT(("left has %d, right has %d\n",
			left->num_tags, right->num_tags));

		if (left->num_tags > 0 && right->num_tags > 0)
		  {
		    /* Reserve the next tag to the right child. */
		    DPRINT(("reserved_tag %d\n", reserved_tag));
		    reserved_tag = next_tag;
		    /* After processing right child. */
		    STACK_PUSHX(stack, next_tag + left->num_tags
				+ right->num_tags);
		    STACK_PUSHX(stack, ADDTAGS_POST_CATENATION);
		    next_tag++;
		  }
		else
		  {
		    /* After processing right child. */
		    STACK_PUSHX(stack, next_tag + left->num_tags
				+ right->num_tags);
		    STACK_PUSHX(stack, ADDTAGS_POST_CATENATION);
		  }

		/* Process right child. */
		STACK_PUSHX(stack, right);
		STACK_PUSHX(stack, ADDTAGS_RECURSE);
		/* After processing left child. */
		STACK_PUSHX(stack, reserved_tag);
		STACK_PUSHX(stack, orig_next_tag + left->num_tags);
		STACK_PUSHX(stack, ADDTAGS_CATENATION_NEXT);
		/* Process left child. */
		STACK_PUSHX(stack, left);
		STACK_PUSHX(stack, ADDTAGS_RECURSE);

	      }
	      break;
	    }
	  break; /* end case: ADDTAGS_RECURSE */

	case ADDTAGS_CATENATION_NEXT:
	  {
	    int reserved_tag;
	    next_tag = (int)stack_pop(stack);
	    reserved_tag = (int)stack_pop(stack);
	    if (reserved_tag >= 0)
	      {
		/* Set the current tag to the reserved tag and move all
		   current registers to it. */
		curr_tag = reserved_tag;
		move_regs(regset, submatch_data, curr_tag);
		if (curr_tag < num_tags)
		  tag_directions[curr_tag] = TAG_MAXIMIZE;
	      }
	    DPRINT(("ast_add_tags: catenation start right, current tag %d, "
		    "next tag %d\n", curr_tag, next_tag));
	  }
	  break;

	case ADDTAGS_POST_CATENATION:
	  next_tag = (int)stack_pop(stack);
	  DPRINT(("ast_add_tags: catenation done, current tag %d, "
		  "next tag %d\n", curr_tag, next_tag));

	  /* Move all registers in the set to the current tag. */
	  move_regs(regset, submatch_data, curr_tag);
	  break;

	case ADDTAGS_POST_ITERATION:
	  /* Move all register in the current set to the reserved tag. */
	  DPRINT(("ast_add_tags: post_iter: current %d\n", curr_tag));
	  move_regs(regset, submatch_data, curr_tag);

	  /* Restore boundary tag value. */
	  curr_boundary_tag = (int)stack_pop(stack);
	  break;

	case ADDTAGS_UNION_NEXT:
	  /* We are rising back up from the left union operand. */
	  {
	    int left_num_tags = (int)stack_pop(stack);
	    tre_ast_node_t *left = stack_pop(stack);

	    if (left_num_tags == 0)
	      {
		/* The marker tag was not added yet.  We must add it here. */
		status = add_tag_right(mem, left, curr_tag);
		regset[0] = -1;
		next_tag++;
	      }

	    curr_tag = (int)stack_pop(stack);

	    DPRINT(("ast_add_tags: union start right, current tag %d, "
		    "next tag %d\n", curr_tag, next_tag));

	    /* Use a new marker tag for the right operand. */
	    curr_marker_tag = curr_tag;
	    moving_marker = 1;

 	    /* Lift the bottom of the `regset' array so that when processing
	       the right operand the items currently in the array are
	       invisible.  The original bottom was saved at ADDTAGS_UNION and
	       will be restored at ADDTAGS_POST_UNION. */
	    while (*regset >= 0)
	      regset++;
	  }
	  break;

	case ADDTAGS_POST_UNION:
	  /* We are rising back up from the right union operand. */
	  {
	    int right_num_tags = (int)stack_pop(stack);
	    tre_ast_node_t *right = stack_pop(stack);
	    next_tag = (int)stack_pop(stack);

	    DPRINT(("ast_add_tags: post_union, curr_tag %d, next_tag %d\n",
		    curr_tag, next_tag));

	    if (right_num_tags == 0)
	      {
		/* The marker tag was not added yet.  We must add it here. */
		status = add_tag_right(mem, right, curr_tag);
		regset[0] = -1;
	      }

	    /* Restore the original bottom of the regset array. */
	    regset = stack_pop(stack);

	    /* Restore curr_marker_tag. */
	    curr_marker_tag = (int)stack_pop(stack);

	    /* To make sure ending tag is not a real tag. */
	    curr_tag = num_tags + 1;

	    /* Lift the remaining tags from both operands outside the union
	       scope. */
	    move_regs(regset, submatch_data, curr_tag);
	  }
	  break;

	} /* end switch(symbol) */
    } /* end while(stack_size(stack) > bottom) */


  myfree(orig_regset);

  DPRINT(("ast_add_tags: Done.  Number of tags %d, ending tag %d.\n",
	  next_tag - 1, curr_tag));
  *end_tag = curr_tag;
  return status;
}



/*
  AST to TNFA compilation routines.
*/

static tre_pos_and_tags_t *
set_empty(tre_mem_t *mem)
{
  tre_pos_and_tags_t *new_set;

  new_set = tre_mem_alloc(mem, sizeof(*new_set));
  if (new_set == NULL)
    return NULL;

  new_set[0].position = -1;
  new_set[0].code_min = -1;
  new_set[0].code_max = -1;
  new_set[0].tags = NULL;
  new_set[0].assertions = 0;
  new_set[0].class = (wctype_t)0;
  new_set[0].neg_classes = NULL;

  return new_set;
}

static tre_pos_and_tags_t *
set_one(tre_mem_t *mem, int position, int code_min, int code_max,
	wctype_t class, wctype_t *neg_classes)
{
  tre_pos_and_tags_t *new_set;

  new_set = tre_mem_alloc(mem, sizeof(*new_set) * 2);
  if (new_set == NULL)
    return NULL;

  new_set[0].position = position;
  new_set[0].code_min = code_min;
  new_set[0].code_max = code_max;
  new_set[0].tags = NULL;
  new_set[0].assertions = 0;
  new_set[0].class = class;
  new_set[0].neg_classes = neg_classes;
  new_set[1].position = -1;
  new_set[1].code_min = -1;
  new_set[1].code_max = -1;
  new_set[1].tags = NULL;
  new_set[1].assertions = 0;
  new_set[1].class = (wctype_t)0;
  new_set[1].neg_classes = NULL;

  return new_set;
}

static tre_pos_and_tags_t *
set_union(tre_mem_t *mem, tre_pos_and_tags_t *set1, tre_pos_and_tags_t *set2,
	  int *tags, int assertions)
{
  int s1, s2, i, j;
  tre_pos_and_tags_t *new_set;
  int *new_tags;
  int num_tags;

  for (num_tags = 0; tags != NULL && tags[num_tags] >= 0; num_tags++);
  for (s1 = 0; set1[s1].position >= 0; s1++);
  for (s2 = 0; set2[s2].position >= 0; s2++);
  new_set = tre_mem_alloc(mem, sizeof(*new_set) * (s1 + s2 + 1));
  if (new_set == NULL)
    return NULL;

  for (s1 = 0; set1[s1].position >= 0; s1++)
    {
      new_set[s1].position = set1[s1].position;
      new_set[s1].code_min = set1[s1].code_min;
      new_set[s1].code_max = set1[s1].code_max;
      new_set[s1].assertions = set1[s1].assertions | assertions;
      new_set[s1].class = set1[s1].class;
      new_set[s1].neg_classes = set1[s1].neg_classes;
      if (set1[s1].tags == NULL && tags == NULL)
	new_set[s1].tags = NULL;
      else
	{
	  for (i = 0; set1[s1].tags != NULL && set1[s1].tags[i] >= 0; i++);
	  new_tags = tre_mem_alloc(mem, (sizeof(*new_tags)
					 * (i + num_tags + 1)));
	  if (new_tags == NULL)
	    return NULL;
	  for (j = 0; j < i; j++)
	    new_tags[j] = set1[s1].tags[j];
	  for (i = 0; i < num_tags; i++)
	    new_tags[j + i] = tags[i];
	  new_tags[j + i] = -1;
	  new_set[s1].tags = new_tags;
	}
    }

  for (s2 = 0; set2[s2].position >= 0; s2++)
    {
      new_set[s1 + s2].position = set2[s2].position;
      new_set[s1 + s2].code_min = set2[s2].code_min;
      new_set[s1 + s2].code_max = set2[s2].code_max;
      new_set[s1 + s2].assertions = set2[s2].assertions;
      new_set[s1 + s2].class = set2[s2].class;
      new_set[s1 + s2].neg_classes = set2[s2].neg_classes;
      if (set2[s2].tags == NULL)
	new_set[s1 + s2].tags = NULL;
      else
	{
	  for (i = 0; set2[s2].tags[i] >= 0; i++);
	  new_tags = tre_mem_alloc(mem, sizeof(*new_tags) * (i + 1));
	  if (new_tags == NULL)
	    return NULL;
	  for (j = 0; j < i; j++)
	    new_tags[j] = set2[s2].tags[j];
	  new_tags[j] = -1;
	  new_set[s1 + s2].tags = new_tags;
	}
    }

  new_set[s1 + s2].position = -1;
  return new_set;
}

/* Finds the empty path through `node' which is the one that should be
   taken according to POSIX.2 rules, and adds the tags on that path to
   `tags'.   `tags' may be NULL.  Returns the number of tags found on the
   path. */
/* XXX - handle stack space exhaustion. */
static int
match_empty(tre_stack_t *stack, tre_ast_node_t *node, int *tags, int *assertions)
{
  tre_literal_t *lit;
  tre_union_t *uni;
  tre_catenation_t *cat;
  tre_iteration_t *iter;
  int i;
  int num_tags_seen = 0;
  int bottom = stack_size(stack);

  stack_push(stack, node);

  while (stack_size(stack) > bottom)
    {
      node = stack_pop(stack);

      switch (node->type)
	{
	case LITERAL:
	  lit = (tre_literal_t *)node->obj;
	  switch (lit->code_min)
	    {
	    case TAG:
	      if (lit->code_max >= 0)
		{
		  if (tags != NULL)
		    {
		      /* Add the tag to `tags'. */
		      for (i = 0; tags[i] >= 0; i++)
			if (tags[i] == lit->code_max)
			  break;
		      if (tags[i] < 0)
			{
			  tags[i] = lit->code_max;
			  tags[i + 1] = -1;
			}
		    }
		  num_tags_seen++;
		}
	      break;
	    case ASSERTION:
	      assert(lit->code_max >= 1
		     || lit->code_max <= ASSERT_LAST);
	      if (assertions != NULL)
		*assertions |= lit->code_max;
	      break;
	    case EMPTY:
	      break;
	    default:
	      assert(0);
	      break;
	    }
	  break;

	case UNION:
	  /* Subexpressions starting earlier take priority over ones
	     starting later, so we prefer the left subexpression over the
	     right subexpression. */
	  uni = (tre_union_t *)node->obj;
	  if (uni->left->nullable)
	    stack_push(stack, uni->left);
	  else if (uni->right->nullable)
	    stack_push(stack, uni->right);
	  else
	    assert(0);
	  break;

	case CATENATION:
	  /* The path must go through both children. */
	  cat = (tre_catenation_t *)node->obj;
	  assert(cat->left->nullable);
	  assert(cat->right->nullable);
	  stack_push(stack, cat->left);
	  stack_push(stack, cat->right);
	  break;

	case ITERATION:
	  /* A match with an empty string is preferred over no match at
	     all, so we go through the argument if possible. */
	  iter = (tre_iteration_t *)node->obj;
	  if (iter->arg->nullable)
	    stack_push(stack, iter->arg);
	  break;

	default:
	  assert(0);
	  break;
	}
    }

  return num_tags_seen;
}


typedef enum {
  NFL_RECURSE,
  NFL_POST_UNION,
  NFL_POST_CATENATION,
  NFL_POST_ITERATION
} tre_nfl_stack_symbol_t;


/* Computes and fills in the fields `nullable', `firstpos', and `lastpos' for
   the nodes of the AST `tree'. */
static reg_errcode_t
ast_compute_nfl(tre_mem_t *mem, tre_stack_t *stack, tre_ast_node_t *tree)
{
  tre_ast_node_t *node;
  tre_nfl_stack_symbol_t symbol;
  reg_errcode_t status = REG_OK;
  int bottom = stack_size(stack);

  STACK_PUSH(stack, tree);
  STACK_PUSH(stack, NFL_RECURSE);

  while (stack_size(stack) > bottom)
    {
      if (status != REG_OK)
	break;
      symbol = (tre_nfl_stack_symbol_t) stack_pop(stack);
      node = stack_pop(stack);
      switch (symbol)
	{
	case NFL_RECURSE:
	  switch (node->type)
	    {
	    case LITERAL:
	      {
		tre_literal_t *lit = (tre_literal_t *)node->obj;
		if (lit->code_min < 0)
		  {
		    /* Tags, empty strings, and zero width assertions:
		       nullable = true, firstpos = {}, and lastpos = {}. */
		    node->nullable = 1;
		    node->firstpos = set_empty(mem);
		    if (node->firstpos == NULL)
		      status = REG_ESPACE;
		    else
		      {
			node->lastpos = set_empty(mem);
			if (node->lastpos == NULL)
			  status = REG_ESPACE;
			else
			  if (lit->code_min == ASSERTION)
			    {
			      node->firstpos->assertions |= lit->code_max;
			      node->lastpos->assertions |= lit->code_max;
			    }
		      }
		  }
		else
		  {
		    /* Literal at position i: nullable = false, firstpos = {i},
		       lastpos = {i}. */
		    node->nullable = 0;
		    node->firstpos = set_one(mem, lit->position, lit->code_min,
					     lit->code_max, 0, NULL);
		    if (node->firstpos == NULL)
		      status = REG_ESPACE;
		    else
		      node->lastpos = set_one(mem, lit->position,
					      lit->code_min, lit->code_max,
					      lit->class, lit->neg_classes);
		    if (node->lastpos == NULL)
		      status = REG_ESPACE;
		  }
		break;
	      }

	    case UNION:
	      /* Compute the attributes for the two subtrees, and after that
		 for this node. */
	      STACK_PUSHX(stack, node);
	      STACK_PUSHX(stack, NFL_POST_UNION);
	      STACK_PUSHX(stack, ((tre_union_t *)node->obj)->right);
	      STACK_PUSHX(stack, NFL_RECURSE);
	      STACK_PUSHX(stack, ((tre_union_t *)node->obj)->left);
	      STACK_PUSHX(stack, NFL_RECURSE);
	      break;

	    case CATENATION:
	      /* Compute the attributes for the two subtrees, and after that
		 for this node. */
	      STACK_PUSHX(stack, node);
	      STACK_PUSHX(stack, NFL_POST_CATENATION);
	      STACK_PUSHX(stack, ((tre_catenation_t *)node->obj)->right);
	      STACK_PUSHX(stack, NFL_RECURSE);
	      STACK_PUSHX(stack, ((tre_catenation_t *)node->obj)->left);
	      STACK_PUSHX(stack, NFL_RECURSE);
	      break;

	    case ITERATION:
	      /* Compute the attributes for the subtree, and after that for
		 this node. */
	      STACK_PUSHX(stack, node);
	      STACK_PUSHX(stack, NFL_POST_ITERATION);
	      STACK_PUSHX(stack, ((tre_iteration_t *)node->obj)->arg);
	      STACK_PUSHX(stack, NFL_RECURSE);
	      break;
	    }
	  break; /* end case: NFL_RECURSE */

	case NFL_POST_UNION:
	  {
	    tre_union_t *uni = (tre_union_t *)node->obj;
	    node->nullable = uni->left->nullable || uni->right->nullable;
	    node->firstpos = set_union(mem, uni->left->firstpos,
				       uni->right->firstpos, NULL, 0);
	    if (node->firstpos == NULL)
		status = REG_ESPACE;
	    else
	      node->lastpos = set_union(mem, uni->left->lastpos,
					uni->right->lastpos, NULL, 0);
	    if (node->lastpos == NULL)
	      status = REG_ESPACE;
	    break;
	  }

	case NFL_POST_ITERATION:
	  {
	    tre_iteration_t *iter = (tre_iteration_t *)node->obj;

	    if (iter->min == 0)
	      node->nullable = 1;
	    else
	      node->nullable = 0;
	    node->firstpos = iter->arg->firstpos;
	    node->lastpos = iter->arg->lastpos;
	    break;
	  }

	case NFL_POST_CATENATION:
	  {
	    int num_tags, *tags, assertions;
	    tre_catenation_t *cat = node->obj;
	    node->nullable = cat->left->nullable && cat->right->nullable;

	    if (cat->left->nullable)
	      {
		num_tags = match_empty(stack, cat->left, NULL, NULL);

		tags = mymalloc(sizeof(int) * (num_tags + 1));
		if (tags == NULL)
		  {
		    status = REG_ESPACE;
		    break;
		  }
		else
		  {
		    tags[0] = -1;
		    assertions = 0;
		    match_empty(stack, cat->left, tags, &assertions);
		    DPRINT(("assertions %d\n", assertions));
		    node->firstpos =
		      set_union(mem, cat->right->firstpos, cat->left->firstpos,
				tags, assertions);

		    myfree(tags);
		    if (node->firstpos == NULL)
		      {
			status = REG_ESPACE;
			break;
		      }
		  }
	      }
	    else
	      {
		node->firstpos = cat->left->firstpos;
	      }

	    if (cat->right->nullable)
	      {
		num_tags = match_empty(stack, cat->right, NULL, NULL);

		tags = mymalloc(sizeof(int) * (num_tags + 1));
		if (tags == NULL)
		  {
		    status = REG_ESPACE;
		    break;
		  }
		else
		  {
		    tags[0] = -1;
		    assertions = 0;
		    match_empty(stack, cat->right, tags, &assertions);
		    node->lastpos =
		      set_union(mem, cat->left->lastpos, cat->right->lastpos,
				tags, assertions);

		    myfree(tags);
		    if (node->lastpos == NULL)
		      {
			status = REG_ESPACE;
			break;
		      }
		  }
	      }
	    else
	      {
		node->lastpos = cat->right->lastpos;
	      }
	    break;
	  }

	default:
	  assert(0);
	  break;
	}
    }

  return status;
}

typedef enum {
  TI_RECURSE,
  TI_POST_CATENATION,
  TI_POST_ITERATION,
  TI_POST_UNION
} ti_stack_symbol_t;

/* Sets the `num_tags' field for each node in `tree' to the number of tags
   needed in the subtree. */
static reg_errcode_t
ast_compute_tag_info(tre_stack_t *stack, tre_ast_node_t *tree)
{
  tre_ast_node_t *node;
  ti_stack_symbol_t symbol;
  reg_errcode_t status = REG_OK;
  int bottom = stack_size(stack);

  /* A flag telling whether the set of submatch registers for the next tag
     is empty or not.  False means empty, true means non-empty.  Initially
     the register set contains the match start point register, so this is
     initially true. */
  int regset = 1;

  STACK_PUSH(stack, tree);
  STACK_PUSH(stack, TI_RECURSE);

  while (stack_size(stack) > bottom)
    {
      if (status != REG_OK)
	break;
      symbol = (ti_stack_symbol_t) stack_pop(stack);
      switch (symbol)
	{
	case TI_RECURSE:
	  node = stack_pop(stack);

	  if (node->submatch_id >= 0)
	    {
	      regset = 1;
	    }

	  switch (node->type)
	    {
	    case LITERAL:
	      node->num_tags = 0;
	      break;
	    case CATENATION:
	      STACK_PUSHX(stack, node);
	      STACK_PUSHX(stack, TI_POST_CATENATION);
	      STACK_PUSHX(stack, ((tre_catenation_t *)node->obj)->left);
	      STACK_PUSHX(stack, TI_RECURSE);
	      STACK_PUSHX(stack, ((tre_catenation_t *)node->obj)->right);
	      STACK_PUSHX(stack, TI_RECURSE);
	      break;
	    case ITERATION:
	      DPRINT(("iteration: tag\n"));
	      node->num_tags++;
	      regset = 0;
	      STACK_PUSHX(stack, node);
	      STACK_PUSHX(stack, TI_POST_ITERATION);
	      STACK_PUSHX(stack, ((tre_iteration_t *)node->obj)->arg);
	      STACK_PUSHX(stack, TI_RECURSE);
	      break;
	    case UNION:
	      if (regset || (node->num_submatches
			     - (node->submatch_id >= 0 ? 1 : 0)) > 0)
		{
		  DPRINT(("union: tag\n"));
		  node->num_tags++;
		}
	      regset = 0;
	      STACK_PUSHX(stack, node);
	      STACK_PUSHX(stack, TI_POST_UNION);
	      STACK_PUSHX(stack, ((tre_catenation_t *)node->obj)->left);
	      STACK_PUSHX(stack, TI_RECURSE);
	      STACK_PUSHX(stack, ((tre_catenation_t *)node->obj)->right);
	      STACK_PUSHX(stack, TI_RECURSE);
	      break;
	    default:
	      assert(0);
	      break;
	    }
	  break;
	case TI_POST_CATENATION:
	  node = stack_pop(stack);
	  node->num_tags += ((tre_catenation_t *)node->obj)->left->num_tags
	    + ((tre_catenation_t *)node->obj)->right->num_tags;
	  break;
	case TI_POST_UNION:
	  {
	    int l = 0, r = 0;
	    node = stack_pop(stack);
	    /* Account for marker tags. */
	    if ((node->num_submatches - (node->submatch_id >= 0 ? 1 : 0)) > 0)
	      {
		if (((tre_union_t *)node->obj)->left->num_tags == 0)
		  l = 1;
		if (((tre_union_t *)node->obj)->right->num_tags == 0)
		  r = 1;
	      }
	    node->num_tags += ((tre_union_t *)node->obj)->left->num_tags
	      + ((tre_union_t *)node->obj)->right->num_tags + l + r;
	  }
	  break;
	case TI_POST_ITERATION:
	  node = stack_pop(stack);
	  node->num_tags += ((tre_iteration_t *)node->obj)->arg->num_tags;
	  break;
	default:
	  assert(0);
	  break;
	}
    }

  return status;
}



typedef struct tnfa_transition tre_tnfa_transition_t;

struct tnfa_transition {
  wint_t code_min;
  wint_t code_max;
  tre_tnfa_transition_t *state;
  int state_id;
  int *tags;
  int assertions;
  wctype_t class;
  wctype_t *neg_classes;
};

typedef struct tnfa_reach tre_tnfa_reach_t;

struct tnfa_reach {
  tre_tnfa_transition_t *state;
  int *tags;
};

typedef struct reach_pos tre_reach_pos_t;

struct reach_pos {
  int pos;
  int **tags;
};

typedef struct tnfa tre_tnfa_t;

struct tnfa {
  tre_tnfa_transition_t *transitions;
  int num_transitions;
  tre_tnfa_transition_t *initial;
  tre_tnfa_transition_t *final;
  tre_submatch_data_t *submatch_data;
  int num_submatches;
  int *tag_directions;
  int *tag_min_bounds;
  int *marker_offs;
  int num_tags;
  int end_tag;
  int num_states;
  int cflags;
};


/* Adds a transition from each position in `p1' to each position in `p2'. */
static reg_errcode_t
make_transitions(tre_pos_and_tags_t *p1, tre_pos_and_tags_t *p2,
		 tre_tnfa_transition_t *transitions,
		 int *counts, int *offs)
{
  tre_pos_and_tags_t *orig_p2 = p2;
  tre_tnfa_transition_t *trans;
  int i, j, k, l, dup, prev_p2_pos;

  if (transitions != NULL)
    while (p1->position >= 0)
      {
	p2 = orig_p2;
	prev_p2_pos = -1;
	while (p2->position >= 0)
	  {
	    if (p2->position == prev_p2_pos)
	      break;
	    prev_p2_pos = p2->position;
	    /* Set `trans' to point to the next unused transition from
	       position `p1->position'. */
	    trans = transitions + offs[p1->position];
	    while (trans->state != NULL)
	      {
#if 0
		/* If we find a previous transition from `p1->position' to
		   `p2->position', it is overwritten.  This can happen only
		   if there are nested loops in the regexp, like in "((a)*)*".
		   In POSIX.2 repetition using the outer loop is always
		   preferred over using the inner loop.  Therefore the
		   transition for the inner loop is useless and can be thrown
		   away. */
		/* XXX - The same position is used for all nodes in a bracket
		   expression, so this optimization cannot be used (it will
		   break bracket expressions) unless I figure out a way to
		   detect it here. */
		if (trans->state_id == p2->position)
		  {
		    DPRINT(("*"));
		    break;
		  }
#endif
		trans++;
	      }
	    if (trans->state == NULL)
	      (trans + 1)->state = NULL;
	    trans->code_min = p1->code_min;
	    trans->code_max = p1->code_max;
	    trans->state = transitions + offs[p2->position];
	    trans->state_id = p2->position;
	    trans->assertions = p1->assertions | p2->assertions
	      | (p1->class ? ASSERT_CHAR_CLASS : 0)
	      | (p1->neg_classes != NULL ? ASSERT_CHAR_CLASS_NEG : 0);
	    trans->class = p1->class;
	    if (p1->neg_classes != NULL)
	      {
		for (i = 0; p1->neg_classes[i] != (wctype_t)0; i++);

		trans->neg_classes =
		  mymalloc(sizeof(*trans->neg_classes) * (i + 1));
		if (trans->neg_classes == NULL)
		  return REG_ESPACE;
		for (i = 0; p1->neg_classes[i] != (wctype_t)0; i++)
		  trans->neg_classes[i] = p1->neg_classes[i];
		trans->neg_classes[i] = (wctype_t)0;
	      }
	    else
	      trans->neg_classes = NULL;

	    i = 0;
	    if (p1->tags != NULL)
	      while(p1->tags[i] >= 0)
		i++;
	    j = 0;
	    if (p2->tags != NULL)
	      while(p2->tags[j] >= 0)
		j++;
	    if (trans->tags != NULL)

	      myfree(trans->tags);

	    trans->tags = mymalloc(sizeof(*trans->tags) * (i + j + 1));
	    DPRINT(("tags = %p\n", trans->tags));
	    if (trans->tags == NULL)
	      return REG_ESPACE;
	    i = 0;
	    if (p1->tags != NULL)
	      while(p1->tags[i] >= 0)
		{
		  trans->tags[i] = p1->tags[i];
		  i++;
		}
	    l = i;
	    j = 0;
	    if (p2->tags != NULL)
	      while (p2->tags[j] >= 0)
		{
		  /* Don't add duplicates. */
		  dup = 0;
		  for (k = 0; k < i; k++)
		    if (trans->tags[k] == p2->tags[j])
		      {
			dup = 1;
			break;
		      }
		  if (!dup)
		    trans->tags[l++] = p2->tags[j];
		  j++;
		}
	    trans->tags[l] = -1;


#ifdef REGEX_DEBUG
	    {
	      int *tags;

	      DPRINT(("from %d to %d on %d-%d", p1->position, p2->position,
		      p1->code_min, p1->code_max));
	      tags = trans->tags;
	      if (tags != NULL)
		{
		  if (*tags >= 0)
		    DPRINT(("/"));
		  while (*tags >= 0)
		    {
		      DPRINT(("%d", *tags));
		      tags++;
		      if (*tags >= 0)
			DPRINT((","));
		    }
		}
	      DPRINT((", assert %d, class %ld, neg_classes %p\n",
		      trans->assertions, (long)trans->class,
		      trans->neg_classes));
	    }
#endif /* REGEX_DEBUG */
	    p2++;
	  }
	p1++;
      }
  else
    /* Compute a maximum limit for the number of transitions leaving
       from each state. */
    while (p1->position >= 0)
      {
	p2 = orig_p2;
	while (p2->position >= 0)
	  {
	    counts[p1->position]++;
	    p2++;
	  }
	p1++;
      }
  return REG_OK;
}

/* Converts the syntax tree to a TNFA.  All the transitions in the TNFA are
   labelled with a single symbol range (there are no transitions on empty
   strings). The TNFA takes O(n^2) space in the worst case, `n' is size of
   the regexp. */
static reg_errcode_t
ast_to_efree_tnfa(tre_ast_node_t *node, tre_tnfa_transition_t *transitions,
		 int *counts, int *offs)
{
  tre_union_t *uni;
  tre_catenation_t *cat;
  tre_iteration_t *iter;
  reg_errcode_t errcode = REG_OK;

  switch (node->type)
    {
    case LITERAL:
      break;
    case UNION:
      uni = (tre_union_t *)node->obj;
      errcode = ast_to_efree_tnfa(uni->left, transitions, counts, offs);
      if (errcode != REG_OK)
	return errcode;
      errcode = ast_to_efree_tnfa(uni->right, transitions, counts, offs);
      break;

    case CATENATION:
      cat = (tre_catenation_t *)node->obj;
      /* Add a transition from each position in cat->left->lastpos
	 to each position in cat->right->firstpos. */
      errcode = make_transitions(cat->left->lastpos, cat->right->firstpos,
				 transitions, counts, offs);
      if (errcode != REG_OK)
	return errcode;
      errcode = ast_to_efree_tnfa(cat->left, transitions, counts, offs);
      if (errcode != REG_OK)
	return errcode;
      errcode = ast_to_efree_tnfa(cat->right, transitions, counts, offs);
      break;

    case ITERATION:
      iter = (tre_iteration_t *)node->obj;
      if (iter->max == -1)
	{
	  /* Add a transition from each position in iter->arg->lastpos
	     to each position in iter->arg->firstpos. */
	  errcode = make_transitions(iter->arg->lastpos, iter->arg->firstpos,
				     transitions, counts, offs);
	  if (errcode != REG_OK)
	    return errcode;
	}
      errcode = ast_to_efree_tnfa(iter->arg, transitions, counts, offs);
      break;
    }
  return errcode;
}


/*
  TNFA simulation routines.
*/

#ifdef REGEX_DEBUG
static void
print_reach(const tre_tnfa_t *tnfa, tre_tnfa_reach_t *reach)
{
  int i;

  while (reach->state != NULL)
    {
      DPRINT((" %p", (void *)reach->state));
      if (tnfa->num_tags > 0)
	{
	  DPRINT(("/"));
	  for (i = 0; i < tnfa->num_tags; i++)
	    {
	      DPRINT(("%d:%d", i, reach->tags[i]));
	      if (i < (tnfa->num_tags-1))
		DPRINT((","));
	    }
	}
      reach++;
    }
  DPRINT(("\n"));

}
#endif /* REGEX_DEBUG */

/* Returns 1 if `t1' wins `t2', 0 otherwise. */
static int
tag_order(int num_tags, int *tag_directions, int *t1, int *t2)
{
  int i;
  /*
  DPRINT(("tag_order\n"));
  DPRINT(("t1[0] = %d, t2[0] = %d\n", *t1, *t2));
  */
  i = 0;
  while (i < num_tags)
    {
      if (tag_directions[i] == TAG_MINIMIZE)
	{
	  DPRINT(("< t1[%d] = %d, t2[%d] = %d\n", i, t1[i], i, t2[i]));
	  if (t1[i] > t2[i])
	    return 0;
	  if (t1[i] < t2[i])
	    return 1;
	}
      else
	{
	  DPRINT(("> t1[%d] = %d, t2[%d] = %d\n", i, t1[i], i, t2[i]));
	  if (t1[i] > t2[i])
	    return 1;
	  if (t1[i] < t2[i])
	    return 0;
	}
      i++;
    }
  /*  assert(0);*/
  return 0;
}

static int
neg_char_classes_match(wctype_t *classes, wint_t wc, int icase)
{
  DPRINT(("neg_char_classes_test: %p, %d, %d\n", classes, wc, icase));
  while (*classes != (wctype_t)0)
    if ((!icase && iswctype(wc, *classes))
	|| (icase && (iswctype(towupper(wc), *classes)
		      || iswctype(towlower(wc), *classes))))
      return 1; /* Match. */
    else
      classes++;
  return 0; /* No match. */
}

#if defined HAVE_MBRTOWC
#define my_mbrtowc(pwc, s, n, ps) (mbrtowc((pwc), (s), (n), (ps)))
#elif defined HAVE_MBTOWC
#define my_mbrtowc(pwc, s, n, ps) (mbtowc((pwc), (s), (n)))
#endif

#define REG_NOTAGS (REG_NOTEOL << 1)

typedef enum { STR_WIDE, STR_BYTE, STR_MBS } tre_str_type_t;

static reg_errcode_t
tnfa_execute(const tre_tnfa_t *tnfa, const void *string, size_t len,
	     tre_str_type_t type, int *match_tags, int eflags,
	     int *match_end_ofs)
{
  wchar_t prev_c = L'\0', next_c = L'\0';
  const char *str_byte = string;
  const wchar_t *str_wide = string;
  char *buf;
  mbstate_t mbstate;
  int pos, num_tags, pos_add = 1, pos_add_next = 0;
  tre_tnfa_transition_t *trans_i;
  tre_tnfa_reach_t *reach, *reach_next, *reach_i, *reach_next_i;
  tre_reach_pos_t *reach_pos;
  int *tag_i;
  int i;
  int reg_notbol = eflags & REG_NOTBOL;
  int reg_noteol = eflags & REG_NOTEOL;
  int reg_newline = tnfa->cflags & REG_NEWLINE;
  int match_eo = -1;       /* end offset of match (-1 if no match found yet) */
  int *tmp_tags = NULL;
  int *tmp_iptr;
  int cflags = tnfa->cflags;

  memset(&mbstate, '\0', sizeof(mbstate));
  if (eflags & REG_NOTAGS)
    num_tags = 0;
  else
    num_tags = tnfa->num_tags;

#define GET_NEXT_WCHAR()                                                      \
  {                                                                           \
    prev_c = next_c;                                                          \
    if (type == STR_BYTE)					      	      \
     next_c = (unsigned char)(*str_byte++);			       	      \
    else if (type == STR_WIDE)						      \
      next_c = *str_wide++;			       			      \
    else if (type == STR_MBS)						      \
      {									      \
        size_t w;	                                                      \
        if (str_byte == NULL)                                                 \
          w = L'\0';                                                          \
        else                                                                  \
          {                                                                   \
            w = my_mbrtowc(&next_c, str_byte, 32, &mbstate);		      \
            if (w < 0)                                                        \
              {	                                                              \
		myfree(buf);						      \
   	        return REG_NOMATCH;                                           \
              }                                                               \
	    pos_add = pos_add_next;                                           \
	    pos_add_next = w;                                                 \
            str_byte += w;                                                    \
          }                                                                   \
     }                                                                        \
  }

  /* XXX - optimize this for 8 bit character sets. */
#define IS_WORD_CHAR(c)  (c == L'_' || iswalnum(c))

#define CHECK_ASSERTIONS(assertions)                                          \
  (assertions != 0							      \
   && (((assertions & ASSERT_AT_BOL)					      \
	&& (pos > 0 || reg_notbol)					      \
	&& (prev_c != L'\n' || !reg_newline))				      \
       || ((assertions & ASSERT_AT_EOL)					      \
	   && (next_c != L'\0' || reg_noteol)				      \
	   && (next_c != L'\n' || !reg_newline))			      \
       || ((assertions & ASSERT_AT_BOW)					      \
	   && (pos > 0 && (IS_WORD_CHAR(prev_c) || !IS_WORD_CHAR(next_c))))   \
       || ((assertions & ASSERT_AT_EOW)                                       \
           && (!IS_WORD_CHAR(prev_c) || IS_WORD_CHAR(next_c)))                \
       || ((assertions & ASSERT_AT_WB)					      \
	   && (pos != 0 && next_c != L'\0'                                    \
	       && IS_WORD_CHAR(prev_c) == IS_WORD_CHAR(next_c)))              \
       || ((assertions & ASSERT_AT_WB_NEG)                                    \
	   && (pos == 0 || next_c == L'\0'                                    \
	       || IS_WORD_CHAR(prev_c) != IS_WORD_CHAR(next_c)))))

  DPRINT(("tnfa_execute, input type %d\n", type));

  /* Allocate memory for temporary data required for matching.  This needs to
     be done for every matching operation to be thread safe.  This allocates
     everything in a single large block from the stack frame using alloca(). */
  {
    int tbytes, rbytes, pbytes, xbytes, total_bytes;
    char *tmp_buf;
    tbytes = sizeof(*tmp_tags) * num_tags;
    tbytes += tbytes % sizeof(long);
    rbytes = sizeof(*reach_next) * (tnfa->num_states + 1);
    rbytes += rbytes % sizeof(long);
    pbytes = sizeof(*reach_pos) * tnfa->num_states;
    pbytes += pbytes % sizeof(long);
    xbytes = sizeof(int) * num_tags;
    xbytes += xbytes % sizeof(long);
    total_bytes = (rbytes + xbytes * tnfa->num_states) * 2 + tbytes + pbytes;

    buf = mymalloc(total_bytes);
    if (buf == NULL)
      return REG_ESPACE;
    memset(buf, 0, total_bytes);
    tmp_tags = (void *)buf;
    tmp_buf = buf + tbytes;
    reach_next = (void *)tmp_buf;
    tmp_buf += rbytes;
    reach = (void *)tmp_buf;
    tmp_buf += rbytes;
    reach_pos = (void *)tmp_buf;
    tmp_buf += pbytes;
    for (i = 0; i < tnfa->num_states; i++)
      {
	reach[i].tags = (void *)tmp_buf;
	tmp_buf += xbytes;
	reach_next[i].tags = (void *)tmp_buf;
	tmp_buf += xbytes;
      }
  }

  for (i = 0; i < tnfa->num_states; i++)
    reach_pos[i].pos = -1;

  reach_next_i = reach_next;
  pos = 0;
  GET_NEXT_WCHAR();

  DPRINT(("pos:chr/code | states and tags\n"));
  DPRINT(("-------------+------------------------------------------------\n"));

  while (1)
    {
      /* If no match found yet, add the initial states to `reach_next'. */
      if (match_eo < 0)
	{
	  DPRINT((" init >"));
	  trans_i = tnfa->initial;
	  while (trans_i->state != NULL)
	    {
	      if (reach_pos[trans_i->state_id].pos < pos)
		{
		  if (CHECK_ASSERTIONS(trans_i->assertions))
		    {
		      DPRINT(("assertion failed\n"));
		      trans_i++;
		      continue;
		    }

		  DPRINT((" %p", (void *)trans_i->state));
		  reach_next_i->state = trans_i->state;
		  for (i = 0; i < num_tags; i++)
		    reach_next_i->tags[i] = -1;
		  tag_i = trans_i->tags;
		  if (tag_i != NULL)
		    while (*tag_i >= 0)
		      {
			reach_next_i->tags[*tag_i] = pos;
			tag_i++;
		      }
		  if (reach_next_i->state == tnfa->final)
		    {
		      match_eo = pos;
		      for (i = 0; i < num_tags; i++)
			match_tags[i] = reach_next_i->tags[i];
		    }
		  reach_pos[trans_i->state_id].pos = pos;
		  reach_pos[trans_i->state_id].tags = &reach_next_i->tags;
		  reach_next_i++;
		}
	      trans_i++;
	    }
	  DPRINT(("\n"));
	  reach_next_i->state = NULL;
	}
      else
	{
	  if (num_tags == 0 || reach_next_i == reach_next)
	    /*We have found a match. */
	    break;
	}

    GET_NEXT_WCHAR();

#ifdef REGEX_DEBUG
    DPRINT(("%3d:%2lc/%05d |", pos, prev_c, (int)prev_c));
    print_reach(tnfa, reach_next);
#endif /* REGEX_DEBUG */

    if (len >= 0)
      {
	if (prev_c == L'\0')
	  break;
      }
    else
      {
	if (pos + pos_add >= len)
	  break;
      }

    /* Swap `reach' and `reach_next'. */
    reach_i = reach;
    reach = reach_next;
    reach_next = reach_i;

    pos += pos_add;

    /* For each state in `reach' see if there is a transition leaving with
       the current input symbol to a state not yet in `reach_next', and
       add the destination states to `reach_next'. */
    reach_i = reach;
    reach_next_i = reach_next;
    while (reach_i->state != NULL)
      {
	trans_i = reach_i->state;
	while (trans_i->state != NULL)
	  {
	    /* Does this transition match the input symbol? */
	    if (trans_i->code_min <= prev_c &&
		trans_i->code_max >= prev_c)
	      {
		if (CHECK_ASSERTIONS(trans_i->assertions)
		    /* Handle character character class transitions. */
		    || ((trans_i->assertions & ASSERT_CHAR_CLASS)
			&& !(cflags & REG_ICASE)
			&& !iswctype((wint_t)prev_c, trans_i->class))
		    || ((trans_i->assertions & ASSERT_CHAR_CLASS)
			&& (cflags & REG_ICASE)
			&& (!iswctype(towlower((wint_t)prev_c), trans_i->class)
			    && !iswctype(towupper((wint_t)prev_c),
					 trans_i->class)))
		    || ((trans_i->assertions & ASSERT_CHAR_CLASS_NEG)
			&& neg_char_classes_match(trans_i->neg_classes,
						  (wint_t)prev_c,
						  cflags & REG_ICASE)))
		  {
		    DPRINT(("assertion failed\n"));
		    trans_i++;
		    continue;
		  }

		/* Compute the tags after this transition. */
		for (i = 0; i < num_tags; i++)
		  tmp_tags[i] = reach_i->tags[i];
		tag_i = trans_i->tags;
		if (tag_i != NULL)
		  while (*tag_i >= 0)
		    {
		      tmp_tags[*tag_i] = pos;
		      tag_i++;
		    }

		if (reach_pos[trans_i->state_id].pos < pos)
		  {
		    /* Found an unvisited node. */
		    reach_next_i->state = trans_i->state;
		    tmp_iptr = reach_next_i->tags;
		    reach_next_i->tags = tmp_tags;
		    tmp_tags = tmp_iptr;
		    reach_pos[trans_i->state_id].pos = pos;
		    reach_pos[trans_i->state_id].tags = &reach_next_i->tags;

		    if (reach_next_i->state == tnfa->final
			&& (match_eo == -1
			    || reach_next_i->tags[0] <= match_tags[0]))
		      {
			match_eo = pos;
			for (i = 0; i < num_tags; i++)
			  match_tags[i] = reach_next_i->tags[i];
		      }
		    reach_next_i++;

		  }
		else
		  {
		    assert(reach_pos[trans_i->state_id].pos == pos);
		    /* Another path has also reached this state.  We choose the
		       winner by examining the tag values for both paths. */
		    if (tag_order(num_tags, tnfa->tag_directions, tmp_tags,
				  *reach_pos[trans_i->state_id].tags))
		      {
			/* The new path wins. */
			tmp_iptr = *reach_pos[trans_i->state_id].tags;
			*reach_pos[trans_i->state_id].tags = tmp_tags;
			if (trans_i->state == tnfa->final
			    && (match_eo == -1
				|| (tmp_tags[0] <= match_tags[0])))
			  {
			    match_eo = pos;
			    for (i = 0; i < num_tags; i++)
			      match_tags[i] = tmp_tags[i];
			  }
			tmp_tags = tmp_iptr;
		      }
		  }
	      }
	    trans_i++;
	  }
	reach_i++;
      }
    reach_next_i->state = NULL;
  }

  DPRINT(("match end offset = %d\n", match_eo));

  *match_end_ofs = match_eo;

  myfree(buf); // REMOVE
  return match_eo >= 0 ? REG_OK : REG_NOMATCH;
}



/*
  Regexp parser.

  The parser is just a simple recursive descent parser for POSIX.2
  regexps.  Supports both the obsolete default syntax and the "extended"
  syntax, and some nonstandard extensions.

*/


/* Characters with special meanings in regexp syntax. */
#define CHAR_PIPE          L'|'
#define CHAR_LPAREN        L'('
#define CHAR_RPAREN        L')'
#define CHAR_LBRACE        L'{'
#define CHAR_RBRACE        L'}'
#define CHAR_LBRACKET      L'['
#define CHAR_RBRACKET      L']'
#define CHAR_MINUS         L'-'
#define CHAR_STAR          L'*'
#define CHAR_QUESTIONMARK  L'?'
#define CHAR_PLUS          L'+'
#define CHAR_PERIOD        L'.'
#define CHAR_COLON         L':'
#define CHAR_EQUAL         L'='
#define CHAR_COMMA         L','
#define CHAR_CARET         L'^'
#define CHAR_DOLLAR        L'$'
#define CHAR_BACKSLASH     L'\\'


static const wchar_t *
expand_macro(const wchar_t *macros[], const wchar_t *regex)
{
  int i;
  for (i = 0; macros[i] != NULL; i += 2)
    if (wcsncmp(macros[i], regex, wcslen(macros[i])) == 0)
      {
	DPRINT(("Expanding macro '%ls' => '%ls'\n", macros[i], macros[i + 1]));
	break;
      }
  if (macros[i] == NULL)
    return NULL;
  return macros[i + 1];
}

/* Some extended (Perl compatible) syntax. */
static const wchar_t *perl_macros[] =
  { L"t", L"\t",             L"n", L"\n",             L"r", L"\r",
    L"f", L"\f",             L"a", L"\a",             L"e", L"\033",
    L"w", L"[[:alnum:]_]",   L"W", L"[^[:alnum:]_]",  L"s", L"[[:space:]]",
    L"S", L"[^[:space:]]",   L"d", L"[[:digit:]]",    L"D", L"[^[:digit:]]",
   NULL };


static reg_errcode_t
new_item(tre_mem_t *mem, int min, int max, int *i, int *max_i,
	 tre_ast_node_t ***items)
{
  reg_errcode_t status;
  tre_ast_node_t **array = *items;
  /* Allocate more space if necessary. */
  if (*i >= *max_i)
    {
      tre_ast_node_t **new_items;
      DPRINT(("out of array space, i = %d\n", *i));
      /* If the array is already 1024 items large, give up -- there's
	 probably an error in the regexp (e.g. not a '\0' terminated
	 string and missing ']') */
      if (*max_i > 1024)
	return REG_ESPACE;
      *max_i *= 2;

      new_items = myrealloc(array, sizeof(*items) * *max_i);
      if (new_items == NULL)
	return REG_ESPACE;
      *items = array = new_items;
    }
  array[*i] = ast_new_literal(mem, min, max, -1);
  status = array[*i] == NULL ? REG_ESPACE : REG_OK;
  (*i)++;
  return status;
}


/* Expands a character class to character ranges. */
static reg_errcode_t
expand_ctype(tre_mem_t *mem, wctype_t class, tre_ast_node_t ***items,
	     int *i, int *max_i, int cflags)
{
  reg_errcode_t status = REG_OK;
  wint_t c;
  int min = -1, max = 0;
  assert(MB_CUR_MAX == 1);

  for (c = 0; c < 256 && status == REG_OK; c++)
    {
      if (iswctype(c, class)
	  || ((cflags & REG_ICASE) && (iswctype(towlower(c), class)
				       || iswctype(towupper(c), class))))
	{
	  if (min < 0)
	    min = c;
	  max = c;
	}
      else if (min >= 0)
	{
	  DPRINT(("range %c (%d) to %c (%d)\n", min, min, max, max));
	  status = new_item(mem, min, max, i, max_i, items);
	  min = -1;
	}
    }
  if (min >= 0)
    status = new_item(mem, min, max, i, max_i, items);
  return status;
}


static int
compare_items(const void *a, const void *b)
{
  tre_ast_node_t *node_a = *(tre_ast_node_t **)a, *node_b = *(tre_ast_node_t **)b;
  tre_literal_t *l_a = node_a->obj, *l_b = node_b->obj;
  int a_min = l_a->code_min, b_min = l_b->code_min;

  if (a_min < b_min)
    return -1;
  else if (a_min > b_min)
    return 1;
  else
    return 0;
}

#define MIN(a,b) ((a) < (b) ? (a) : (b))
#define MAX(a,b) ((a) > (b) ? (a) : (b))


/* Maximum number of character classes that can occur in a negated bracket
   expression. */
#define MAX_NEG_CLASSES 64

static reg_errcode_t
parse_bracket_items(tre_mem_t *mem, const wchar_t **re, const wchar_t *re_end,
		    int *position, int cflags, int negate,
		    wctype_t neg_classes[], int *num_neg_classes,
		    tre_ast_node_t ***items, int *num_items, int *items_size)
{
  const wchar_t *regex = *re;
  reg_errcode_t status = REG_OK;
  wctype_t class = (wctype_t)0;
  int i = *num_items;
  int max_i = *items_size;
  int skip;

  /* Build an array of the items in the bracket expression. */
  while (status == REG_OK)
    {
      skip = 0;
      if (regex == re_end)
	  status = REG_EBRACK;
      else if (*regex == CHAR_RBRACKET && regex > *re)
	{
	  DPRINT(("parse_bracket:   done: '%*ls'\n", re_end - regex, regex));
	  regex++;
	  break;
	}
      else
	{
	  int min, max;

	  class = (wctype_t)0;
	  if (*(regex + 1) == CHAR_MINUS && *(regex + 2) != CHAR_RBRACKET)
	    {
	      DPRINT(("parse_bracket:  range: '%*ls'\n",
		      re_end - regex, regex));
	      min = *regex;
	      max = *(regex + 2);
	      regex += 3;
	      /* XXX - Should use collation order instead of encoding values
		 in character ranges. */
	      if (min > max)
		status = REG_ERANGE;
	    }
	  else if (*regex == CHAR_LBRACKET && *(regex + 1) == CHAR_PERIOD)
	    status = REG_ECOLLATE;
	  else if (*regex == CHAR_LBRACKET && *(regex + 1) == CHAR_EQUAL)
	    status = REG_ECOLLATE;
	  else if (*regex == CHAR_LBRACKET && *(regex + 1) == CHAR_COLON)
	    {
	      wchar_t tmp_wcs[64];
	      char tmp_str[64];
	      wchar_t *endptr;
	      int len;
	      DPRINT(("parse_bracket:  class: '%*ls'\n",
		      re_end - regex, regex));
	      endptr = wcschr(regex + 2, CHAR_COLON);
	      if (endptr != NULL)
		{
		  len = MIN(endptr - regex - 2, 63);
		  wcsncpy(tmp_wcs, regex + 2, len);
		  tmp_wcs[len] = L'\0';
#if defined HAVE_WCSRTOMBS
                  {
		    mbstate_t state;
		    const wchar_t *src = tmp_wcs;
		    memset(&state, '\0', sizeof(state));
		    len = wcsrtombs(tmp_str, &src, 63, &state);
		  }
#elif defined HAVE_WCSTOMBS
		  len = wcstombs(tmp_str, tmp_wcs, 63);
#endif
		  tmp_str[len] = '\0';
		  class = wctype(tmp_str);
		  if (!class)
		    status = REG_ECTYPE;
		  /* Optimize character classes for 8 bit character sets. */
		  if (status == REG_OK && MB_CUR_MAX == 1)
		    {
		      status =
			expand_ctype(mem, class, items, &i, &max_i, cflags);
		      class = (wctype_t)0;
		      skip = 1;
		    }
		  regex = endptr + 2;
		}
	      else
		status = REG_ECTYPE;
	      min = 0;
	      max = WCHAR_MAX;
	    }
	  else
	    {
	      DPRINT(("parse_bracket:   char: '%*ls'\n", re_end - regex,
		      regex));
	      if (*regex == CHAR_MINUS && *(regex + 1) != CHAR_RBRACKET
		  && *re != regex)
		/* Two ranges are not allowed to share and endpoint. */
		status = REG_ERANGE;
	      min = max = *regex++;
	    }

	  if (status != REG_OK)
	    break;

	  if (class && negate)
	    if (*num_neg_classes >= MAX_NEG_CLASSES)
	      status = REG_ESPACE;
	    else
	      neg_classes[(*num_neg_classes)++] = class;
	  else if (!skip)
	    {
	      status = new_item(mem, min, max, &i, &max_i, items);
	      if (status != REG_OK)
		break;
	      ((tre_literal_t*)((*items)[i-1])->obj)->class = class;
	    }

	  /* Add opposite-case counterpoints if REG_ICASE is present.
	     This is broken if there are more than two "same" characters. */
	  if (cflags & REG_ICASE && !class && status == REG_OK && !skip)
	    {
	      int cmin, ccurr;

	      DPRINT(("adding opposite-case counterpoints\n"));
	      while (min <= max)
		{
		  if (iswlower(min))
		    {
		      cmin = ccurr = towupper(min++);
		      while (iswlower(min) && towupper(min) == ccurr + 1
			     && min <= max)
			ccurr = towupper(min++);
		      status = new_item(mem, cmin, ccurr, &i, &max_i, items);
		    }
		  else if (iswupper(min))
		    {
		      cmin = ccurr = towlower(min++);
		      while (iswupper(min) && towlower(min) == ccurr + 1
			     && min <= max)
			ccurr = towlower(min++);
		      status = new_item(mem, cmin, ccurr, &i, &max_i, items);
		    }
		  else min++;
		  if (status != REG_OK)
		    break;
		}
	      if (status != REG_OK)
		break;
	    }
	}
    }
  *num_items = i;
  *items_size = max_i;
  *re = regex;
  return status;
}

static reg_errcode_t
parse_bracket(tre_mem_t *mem, const wchar_t **re, const wchar_t *re_end,
	      tre_ast_node_t **result, int *position, int cflags)
{
  const wchar_t *regex = *re;
  tre_ast_node_t *node = NULL;
  int negate = 0;
  reg_errcode_t status = REG_OK;
  tre_ast_node_t **items, *u, *n;
  int i = 0, j, max_i = 32, curr_max, curr_min;
  wctype_t neg_classes[MAX_NEG_CLASSES];
  int num_neg_classes = 0;

  /* Start off with an array of `max_i' elements. */

  items = mymalloc(sizeof(*items) * max_i);
  if (items == NULL)
    return REG_ESPACE;

  if (*regex == CHAR_CARET)
    {
      DPRINT(("parse_bracket: negate: '%*ls'\n", re_end - regex, regex));
      negate = 1;
      regex++;
      (*re)++;
    }

  status = parse_bracket_items(mem, re, re_end, position, cflags, negate,
			       neg_classes, &num_neg_classes, &items,
			       &i, &max_i);

  if (status != REG_OK)
    goto parse_bracket_done;

  /* Sort the array if we need to negate it. */
  if (negate)
    qsort(items, i, sizeof(*items), compare_items);

  curr_max = curr_min = 0;
  /* Build a union of the items in the array, negated if necessary. */
  for (j = 0; j < i && status == REG_OK; j++)
    {
      int min, max;
      tre_literal_t *l = items[j]->obj;
      min = l->code_min;
      max = l->code_max;

      DPRINT(("item: %d - %d, class %ld, curr_max = %d\n",
	      (int)l->code_min, (int)l->code_max, (long)l->class, curr_max));

      if (negate)
	{
	  if (min < curr_max)
	    {
	      /* Overlap. */
	      curr_max = MAX(max + 1, curr_max);
	      DPRINT(("overlap, curr_max = %d\n", curr_max));
	      l = NULL;
	    }
	  else
	    {
	      /* No overlap. */
	      curr_max = min - 1;
	      if (curr_max >= curr_min)
		{
		  DPRINT(("no overlap\n"));
		  l->code_min = curr_min;
		  l->code_max = curr_max;
		}
	      else
		{
		  DPRINT(("no overlap, zero room\n"));
		  l = NULL;
		}
	      curr_min = curr_max = max + 1;
	    }
	}

      if (l != NULL)
	{
	  int k;
	  DPRINT(("creating %d - %d\n", (int)l->code_min, (int)l->code_max));
	  l->position = *position;
	  if (num_neg_classes > 0)
	    {
	      l->neg_classes =
		tre_mem_alloc(mem, (sizeof(l->neg_classes)
				    * (num_neg_classes + 1)));
	      if (l->neg_classes == NULL)
		{
		  status = REG_ESPACE;
		  break;
		}
	      for (k = 0; k < num_neg_classes; k++)
		l->neg_classes[k] = neg_classes[k];
	      l->neg_classes[k] = (wctype_t)0;
	    }
	  else
	    l->neg_classes = NULL;
	  if (node == NULL)
	    node = items[j];
	  else
	    {
	      u = ast_new_union(mem, node, items[j]);
	      if (u == NULL)
		status = REG_ESPACE;
	      node = u;
	    }
	}
    }

  if (status != REG_OK)
    goto parse_bracket_done;

  if (negate)
    {
      int k;
      DPRINT(("final: creating %d - %d\n", curr_min, (int)WCHAR_MAX));
      n = ast_new_literal(mem, curr_min, WCHAR_MAX, *position);
      if (n == NULL)
	status = REG_ESPACE;
      else
	{
	  tre_literal_t *l = n->obj;
	  if (num_neg_classes > 0)
	    {
	      l->neg_classes =
		tre_mem_alloc(mem, (sizeof(l->neg_classes)
				    * (num_neg_classes + 1)));
	      if (l->neg_classes == NULL)
		{
		  status = REG_ESPACE;
		  goto parse_bracket_done;
		}
	      for (k = 0; k < num_neg_classes; k++)
		l->neg_classes[k] = neg_classes[k];
	      l->neg_classes[k] = (wctype_t)0;
	    }
	  else
	    l->neg_classes = NULL;
	  if (node == NULL)
	    node = n;
	  else
	    {
	      u = ast_new_union(mem, node, n);
	      if (u == NULL)
		status = REG_ESPACE;
	      node = u;
	    }
	}
    }

  if (status != REG_OK)
    goto parse_bracket_done;

#ifdef REGEX_DEBUG
  ast_print(node);
#endif /* REGEX_DEBUG */

 parse_bracket_done:

  myfree(items);
  (*position)++;
  *result = node;
  return status;
}


#define REG_NOFIRSTSUB  (REG_NOSUB << 1)

static reg_errcode_t
parse_re(tre_mem_t *mem, tre_stack_t *stack, tre_ast_node_t **root_node,
	 const wchar_t *regex, int len, int *submatch_id, int *position,
	 int cflags);

static reg_errcode_t
parse_bound(tre_mem_t *mem, tre_stack_t *stack, const wchar_t **regex,
	    const wchar_t *regex_end, const wchar_t *atom, int submatch_id,
	    tre_ast_node_t **result, int *position, int cflags)
{
  int min, max;
  const wchar_t *r = *regex;
  const wchar_t *r_end;
  tre_ast_node_t *non_nullables = NULL, *nullables = NULL;
  tre_ast_node_t *x = NULL, *tmp_node = NULL;

  /* Find the ending '}'. */
  r_end = *regex;
  while (r_end < regex_end && *r_end != CHAR_RBRACE)
    r_end++;

  if (r_end == regex_end)
    return REG_EBRACE;

  min = -1;
  while (*r >= L'0' && *r <= L'9')
    {
      if (min < 0)
	min = 0;
      min = 10 * min + *r - L'0';
      r++;
    }
  max = min;

  if (*r == CHAR_COMMA)
    {
      r++;
      if ((cflags & REG_EXTENDED && *r == CHAR_RBRACE)
	  || (!(cflags & REG_EXTENDED) && *r == CHAR_BACKSLASH
	      && *(r+1) == CHAR_RBRACE))
	max = -1;
      else
	{
	  max = 0;
	  while (*r >= L'0' && *r <= L'9')
	    {
	      max = 10 * max + *r - L'0';
	      r++;
	    }
	}
    }

  if ((cflags & REG_EXTENDED && *r != CHAR_RBRACE)
      || (!(cflags & REG_EXTENDED)
	  && (*r != CHAR_BACKSLASH
	      || *(r+1) != CHAR_RBRACE)))
    return REG_BADBR;

  r++;
  if (!(cflags & REG_EXTENDED))
    r++;

  if (min < 0 || (max >= 0 && min > max) || max > RE_DUP_MAX)
    return REG_BADBR;

  DPRINT(("parse_bound: min %d, max %d\n", min, max));

  if (min > 1 || max > 1)
    {
      int i, sub;
      DPRINT(("parse_bound: expanding to %d non-nullable, %d nullable, "
	      "position %d\n", min, max - min, *position));
      if (min > 0)
	{
	  non_nullables = *result;
	  *result = NULL;
	}
      else
	non_nullables = NULL;
      for (i = 1; i < min; i++)
	{
	  reg_errcode_t status;
	  sub = submatch_id;
	  x = NULL;
	  tmp_node = NULL;
	  status = parse_re(mem, stack, &x, atom, *regex - atom - 1, &sub,
			    position, cflags | REG_NOFIRSTSUB);
	  if (status != REG_OK)
	    return status;
	  if (non_nullables == NULL)
	    non_nullables = x;
	  else
	    {
	      tmp_node = ast_new_catenation(mem, x, non_nullables);
	      if (tmp_node == NULL)
		return REG_ESPACE;
	      non_nullables = tmp_node;
	    }
	}

      nullables = NULL;
      if (non_nullables == NULL)
	{
	  tmp_node = ast_new_literal(mem, EMPTY, -1, -1);
	  if (tmp_node == NULL)
	    return REG_ESPACE;
	  nullables = ast_new_union(mem, tmp_node, *result);
	  max--;
	  if (nullables == NULL)
	    return REG_ESPACE;
	  *result = NULL;
	}
      if (max - min < 0)
	{
	  reg_errcode_t status;
	  status = parse_re(mem, stack, &tmp_node, atom, *regex - atom - 1,
			    &sub, position, cflags | REG_NOFIRSTSUB);
	  if (status != REG_OK)
	    return status;
	  nullables = ast_new_iteration(mem, tmp_node, 0, -1);
	  if (nullables == NULL)
	    return REG_ESPACE;
	}
      else
	for (i = 0; i < max - min; i++)
	  {
	    reg_errcode_t status;
	    x = NULL;
	    tmp_node = NULL;
	    sub = submatch_id;
	    status = parse_re(mem, stack, &x, atom, *regex - atom - 1, &sub,
			      position, cflags | REG_NOFIRSTSUB);
	    if (status != REG_OK)
	      return status;
	    if (nullables == NULL)
	      {
		tmp_node = ast_new_literal(mem, EMPTY, -1, -1);
		if (tmp_node == NULL)
		  return REG_ESPACE;
		nullables = ast_new_union(mem, x, tmp_node);
		if (nullables == NULL)
		  return REG_ESPACE;
		x = NULL;
		tmp_node = NULL;
	      }
	    else
	      {
		tmp_node = ast_new_catenation(mem, x, nullables);
		if (tmp_node == NULL)
		  return REG_ESPACE;
		nullables = NULL;
		x = ast_new_literal(mem, EMPTY, -1, -1);
		if (x == NULL)
		  return REG_ESPACE;
		nullables = ast_new_union(mem, x, tmp_node);
		if (nullables == NULL)
		  return REG_ESPACE;
		x = NULL;
		tmp_node = NULL;
	      }
	  }
      if (non_nullables == NULL)
	*result = nullables;
      else if (nullables == NULL)
	*result = non_nullables;
      else
	*result = ast_new_catenation(mem, non_nullables, nullables);
      if (*result == NULL)
	return REG_ESPACE;
    }
  else if (min == 0 && max == 0)
    {
      *result = ast_new_literal(mem, EMPTY, -1, -1);
      if (*result == NULL)
	return REG_ESPACE;
    }
  else
    {
      tmp_node = ast_new_iteration(mem, *result, min, max);
      if (tmp_node == NULL)
	return REG_ESPACE;
      *result = tmp_node;
    }

  *regex = r;
  return REG_OK;
}

typedef enum {
  PARSE_RE = 0,
  PARSE_ATOM,
  PARSE_MARK_FOR_SUBMATCH,
  PARSE_BRANCH,
  PARSE_PIECE,
  PARSE_CATENATION,
  PARSE_POST_CATENATION,
  PARSE_UNION,
  PARSE_POST_UNION,
  PARSE_POSTFIX
} tre_parse_re_stack_symbol_t;

/* Parses a wide character regexp pattern into a syntax tree.
   Having to support two kinds of REs makes things ugly. */
static reg_errcode_t
parse_re(tre_mem_t *mem, tre_stack_t *stack, tre_ast_node_t **root_node,
	 const wchar_t *regex, int len, int *submatch_id, int *position,
	 int cflags)
{
  tre_ast_node_t *result = NULL;
  tre_parse_re_stack_symbol_t symbol;
  reg_errcode_t status = REG_OK;
  int bottom = stack_size(stack);
  const wchar_t *prev_atom = NULL;
  int atom_smid = 0, depth = 0;
  const wchar_t *regex_end = regex + len;

  DPRINT(("parse_re: parsing '%*ls', len = %d\n", len, regex, len));

  if (!(cflags & REG_NOFIRSTSUB))
    {
      STACK_PUSH(stack, *submatch_id);
      STACK_PUSH(stack, PARSE_MARK_FOR_SUBMATCH);
      (*submatch_id)++;
    }
  STACK_PUSH(stack, PARSE_RE);

  /* The following is basically just a recursive descent parser.  I use
     an explicit stack instead of recursive functions mostly because of
     two reasons: compatibility with systems which have an overflowable
     call stack, and efficiency (both in lines of code and speed).  */
  while (stack_size(stack) > bottom && status == REG_OK)
    {
      if (status != REG_OK)
	break;
      symbol = (tre_parse_re_stack_symbol_t)stack_pop(stack);
      switch (symbol)
	{
	case PARSE_RE:
	  /* Parse a full regexp.  A regexp is one or more branches,
	     separated by the union operator `|'. */
	  if (cflags & REG_EXTENDED)
	    STACK_PUSHX(stack, PARSE_UNION);
	  STACK_PUSHX(stack, PARSE_BRANCH);
	  break;

	case PARSE_BRANCH:
	  /* Parse a branch.  A branch is one or more pieces, concatenated.
             A piece is an atom possibly followed by a postfix operator. */
	  STACK_PUSHX(stack, PARSE_CATENATION);
	  STACK_PUSHX(stack, PARSE_PIECE);
	  break;

	case PARSE_PIECE:
	  /* Parse a piece.  A piece is an atom possibly followed by one
	     or more postfix operators. */
	  STACK_PUSHX(stack, PARSE_POSTFIX);
	  STACK_PUSHX(stack, PARSE_ATOM);
	  break;

	case PARSE_CATENATION:
	  /* If the expression has not ended, parse another piece. */
	  {
	    wchar_t c;
	    if (regex == regex_end)
	      break;
	    c = *regex;
	    if (cflags & REG_EXTENDED && c == CHAR_PIPE)
	      break;
	    if ((cflags & REG_EXTENDED && c == CHAR_RPAREN && depth > 0)
		|| (!(cflags & REG_EXTENDED)
		    && (c == CHAR_BACKSLASH
			&& *(regex + 1) == CHAR_RPAREN)))
	      {
		if (!(cflags & REG_EXTENDED) && depth == 0)
		  status = REG_EPAREN;
		DPRINT(("parse_re:   group end: '%*ls'\n",
			regex_end - regex, regex));
		depth--;
		if (!(cflags & REG_EXTENDED))
		  regex += 2;
		break;
	      }

	    STACK_PUSHX(stack, PARSE_CATENATION);
	    STACK_PUSHX(stack, result);
	    STACK_PUSHX(stack, PARSE_POST_CATENATION);
	    STACK_PUSHX(stack, PARSE_PIECE);
	    break;
	  }

	case PARSE_POST_CATENATION:
	  {
	    tre_ast_node_t *tree = stack_pop(stack);
	    tre_ast_node_t *tmp_node = ast_new_catenation(mem, tree, result);
	    if (tmp_node == NULL)
	      return REG_ESPACE;
	    result = tmp_node;
	    break;
	  }

	case PARSE_UNION:
	  switch (*regex)
	    {
	    case CHAR_PIPE:
	      DPRINT(("parse_re:       union: '%*ls'\n",
		      regex_end - regex, regex));
	      STACK_PUSHX(stack, PARSE_UNION);
	      STACK_PUSHX(stack, result);
	      STACK_PUSHX(stack, PARSE_POST_UNION);
	      STACK_PUSHX(stack, PARSE_BRANCH);
	      regex++;
	      break;

	    case CHAR_RPAREN:
	      regex++;
	      break;

	    default:
	      break;
	    }
	  break;

	case PARSE_POST_UNION:
	  {
	    tre_ast_node_t *tmp_node;
	    tre_ast_node_t *tree = stack_pop(stack);
	    tmp_node = ast_new_union(mem, tree, result);
	    if (tmp_node == NULL)
	      return REG_ESPACE;
	    result = tmp_node;
	    break;
	  }

	case PARSE_POSTFIX:
	  /* Parse postfix operators. */
	  if (regex == regex_end)
	    break;
	  switch (*regex)
	    {
	    case CHAR_STAR:
	      {
		tre_ast_node_t *tmp_node;

		DPRINT(("parse_re:        star: '%*ls'\n",
			regex_end - regex, regex));
		regex++;
		tmp_node = ast_new_iteration(mem, result, 0, -1);
		if (tmp_node == NULL)
		  return REG_ESPACE;
		result = tmp_node;
		STACK_PUSHX(stack, PARSE_POSTFIX);
		break;
	      }

	    case CHAR_BACKSLASH:
	      if (!(cflags & REG_EXTENDED)
		  && *(regex + 1) == CHAR_LBRACE)
		regex++;
	      else
		break;

	    case CHAR_LBRACE:
	      DPRINT(("parse_re:       bound: '%*ls'\n",
		      regex_end - regex, regex));
	      regex++;

	      status = parse_bound(mem, stack, &regex, regex_end, prev_atom,
				   atom_smid, &result, position, cflags);
	      if (status != REG_OK)
		return status;
	      STACK_PUSHX(stack, PARSE_POSTFIX);
	      break;

	    case CHAR_PLUS:
	      if (cflags & REG_EXTENDED)
		{
		  tre_ast_node_t *tmp_node;

		  DPRINT(("parse_re:        plus: '%*ls'\n",
			  regex_end - regex, regex));
		  regex++;
		  tmp_node = ast_new_iteration(mem, result, 1, -1);
		  if (tmp_node == NULL)
		    return REG_ESPACE;
		  result = tmp_node;
		  STACK_PUSHX(stack, PARSE_POSTFIX);
		}
	      break;

	    case CHAR_QUESTIONMARK:
	      if (cflags & REG_EXTENDED)
		{
		  tre_ast_node_t *tmp_node;
		  DPRINT(("parse_re:         opt: '%*ls'\n",
			  regex_end - regex, regex));
		  regex++;
		  tmp_node = ast_new_iteration(mem, result, 0, 1);
		  if (tmp_node == NULL)
		    return REG_ESPACE;
		  result = tmp_node;
		  STACK_PUSHX(stack, PARSE_POSTFIX);
		}
	      break;
	    }
	  break;

	case PARSE_ATOM:
	  /* Parse an atom.  An atom is a regular expression enclosed in `()',
	     an empty set of `()', a bracket expression, `.', `^', `$',
	     a `\' followed by a character, or a single character. */
	  prev_atom = regex;
	  atom_smid = *submatch_id;
	  switch (*regex)
	    {
	    case CHAR_LPAREN:  /* parenthesized subexpression */
	      if (cflags & REG_EXTENDED
		  || *(regex - 1) == CHAR_BACKSLASH)
		{
		  DPRINT(("parse_re: group begin: '%*ls', submatch id %d\n",
			  regex_end - regex, regex, *submatch_id));
		  depth++;
		  regex++;
		  /* First parse a whole RE, then mark the resulting tree
		     for submatching. */
		  STACK_PUSHX(stack, *submatch_id);
		  STACK_PUSHX(stack, PARSE_MARK_FOR_SUBMATCH);
		  STACK_PUSHX(stack, PARSE_RE);
		  (*submatch_id)++;
		}
	      else
		goto parse_literal;
	      break;

	    case CHAR_RPAREN:  /* end of current subexpression */
	      if ((cflags & REG_EXTENDED && depth > 0)
		  || *(regex - 1) == CHAR_BACKSLASH)
		{
		  DPRINT(("parse_re:       empty: '%*ls'\n",
			  regex_end - regex, regex));
		  /* We were expecting an atom, but instead the current
		     subexpression was closed.  POSIX leaves the meaning of
		     this to be implementation-defined.  We interpret this as
		     an empty expression (which matches an empty string).  */
		  result = ast_new_literal(mem, EMPTY, -1, -1);
		  if (result == NULL)
		    return REG_ESPACE;
		  if (!(cflags & REG_EXTENDED))
		    regex--;
		}
	      else
		goto parse_literal;
	      break;

	    case CHAR_LBRACKET: /* bracket expression */
	      DPRINT(("parse_re:     bracket: '%*ls'\n",
		      regex_end - regex, regex));
	      regex++;
	      status = parse_bracket(mem, &regex, regex_end, &result,
				     position, cflags);
	      if (status != REG_OK)
		return status;
	      break;

	    case CHAR_BACKSLASH:
	      if (!(cflags & REG_EXTENDED)
		  && (*(regex + 1) == CHAR_LPAREN
		      || *(regex + 1) == CHAR_RPAREN))
		{
		  /* Chew off the backslash and try again. */
		  regex++;
		  STACK_PUSHX(stack, PARSE_ATOM);
	        }
	      else
		{
		  const wchar_t *r = expand_macro(perl_macros, regex + 1);
		  if (r != NULL)
		    {
		      /* nonstandard syntax extension */
		      status = parse_re(mem, stack, &result, r, wcslen(r),
					submatch_id, position,
					cflags | REG_NOFIRSTSUB);
		      if (status != REG_OK)
			return status;
		      regex += 2;
		    }
		  else
		    {
		      regex++;
		      if (regex == regex_end)
			/* Trailing backslash. */
			return REG_EESCAPE;
		      switch (*regex)
			{
			case L'b':
			  result = ast_new_literal(mem, ASSERTION,
						   ASSERT_AT_WB, -1);
			  break;
			case L'B':
			  result = ast_new_literal(mem, ASSERTION,
						   ASSERT_AT_WB_NEG, -1);
			  break;
			case L'<':
			  result = ast_new_literal(mem, ASSERTION,
						   ASSERT_AT_BOW, -1);
			  break;
			case L'>':
			  result = ast_new_literal(mem, ASSERTION,
						   ASSERT_AT_EOW, -1);
			  break;
			default:
			  /* escaped character */
			  DPRINT(("parse_re:     escaped: '%*ls'\n",
				  regex_end - regex + 1, regex - 1));
			  result = ast_new_literal(mem, *regex, *regex,
						   *position);
			  (*position)++;
			  break;
			}
		      if (result == NULL)
			return REG_ESPACE;
		      regex++;
		    }
		}
	      break;

	    case CHAR_PERIOD:    /* the any-symbol */
	      DPRINT(("parse_re:         any: '%*ls'\n",
		      regex_end - regex, regex));
	      if (cflags & REG_NEWLINE)
		{
		  tre_ast_node_t *tmp_node1;
		  tre_ast_node_t *tmp_node2;
		  tmp_node1 = ast_new_literal(mem, 0, L'\n' - 1, *position);
		  if (tmp_node1 == NULL)
		    return REG_ESPACE;
		  tmp_node2 = ast_new_literal(mem, L'\n' + 1, WCHAR_MAX,
					      *position + 1);
		  if (tmp_node2 == NULL)
		    return REG_ESPACE;
		  result = ast_new_union(mem, tmp_node1, tmp_node2);
		  if (result == NULL)
		    return REG_ESPACE;
		  (*position) += 2;
		}
	      else
		{
		  result = ast_new_literal(mem, 0, WCHAR_MAX, *position);
		  if (result == NULL)
		    return REG_ESPACE;
		  (*position)++;
		}
	      regex++;
	      break;

	    case CHAR_CARET:     /* beginning of line anchor */
	      if (cflags & REG_EXTENDED || *position == 0
		  || *(regex - 1) == CHAR_LPAREN)
		{
		  DPRINT(("parse_re:  line begin: '%*ls'\n",
			  regex_end - regex, regex));
		  result = ast_new_literal(mem, ASSERTION, ASSERT_AT_BOL, -1);
		  if (result == NULL)
		    return REG_ESPACE;
		  regex++;
		}
	      else
		goto parse_literal;
	      break;

	    case CHAR_DOLLAR:    /* end of line anchor */
	      if (cflags & REG_EXTENDED
		  || (regex + 2 < regex_end
		      && *(regex + 1) == CHAR_BACKSLASH
		      && *(regex + 2) == CHAR_RPAREN)
		  || regex + 1 == regex_end)
		{
		  DPRINT(("parse_re:    line end: '%*ls'\n",
			  regex_end - regex, regex));
		  result = ast_new_literal(mem, ASSERTION, ASSERT_AT_EOL, -1);
		  if (result == NULL)
		    return REG_ESPACE;
		  regex++;
		}
	      else
		goto parse_literal;
	      break;

	    default:
	    parse_literal:
	      if (regex == regex_end)
		return REG_EPAREN;

	      DPRINT(("parse_re:     literal: '%*ls'\n",
		      regex_end - regex, regex));
	      /* Note that we can't use an iswalpha() test here, since there
		 are characters which are alphabetic but neither upper or
		 lower case. */
	      if (cflags & REG_ICASE
		  && (iswupper(*regex) || iswlower(*regex)))
		{
		  tre_ast_node_t *tmp_node1;
		  tre_ast_node_t *tmp_node2;

		  /* XXX - Can there be more than one opposite-case
		     counterpoints for some character in some locale?  Or
		     more than two characters which all should be regarded
		     the same character if case is ignored?  If yes, there
		     does not seem to be a portable way to detect it.  I guess
		     that at least for multi-character collating elements there
		     could be several opposite-case counterpoints, but they
		     cannot be supported portably anyway. */
		  tmp_node1 = ast_new_literal(mem, towupper(*regex),
					      towupper(*regex), *position);
		  if (tmp_node1 == NULL)
		    return REG_ESPACE;
		  tmp_node2 = ast_new_literal(mem, towlower(*regex),
					      towlower(*regex), *position);
		  if (tmp_node2 == NULL)
		    return REG_ESPACE;
		  result = ast_new_union(mem, tmp_node1, tmp_node2);
		  if (result == NULL)
		    return REG_ESPACE;
		}
	      else
		{
		  result = ast_new_literal(mem, *regex, *regex, *position);
		  if (result == NULL)
		    return REG_ESPACE;
		}
	      (*position)++;
	      regex++;
	      break;
	    }
	  break;

	case PARSE_MARK_FOR_SUBMATCH:
	  {
	    int submatch_id = (int)stack_pop(stack);

	    if (result->submatch_id >= 0)
	      {
		tre_ast_node_t *n, *tmp_node;
		n = ast_new_literal(mem, EMPTY, -1, -1);
		if (n == NULL)
		  return REG_ESPACE;
		tmp_node = ast_new_catenation(mem, n, result);
		if (tmp_node == NULL)
		  return REG_ESPACE;
		tmp_node->num_submatches = result->num_submatches;
		result = tmp_node;
	      }
	    result->submatch_id = submatch_id;
	    result->num_submatches++;
	    break;
	  }
	}
    }

  if (status == REG_OK)
    *root_node = result;

  return status;
}



/*
   POSIX.2 and wide character API
*/

#define ERROR_EXIT(err)           \
 do {                             \
     errcode = err;               \
     goto error_exit;             \
   }                              \
 while (0);

int
regwncomp(regex_t *preg, const wchar_t *regex, size_t n, int cflags)
{
  tre_stack_t *stack;
  tre_ast_node_t *tree, *tmp_ast_l, *tmp_ast_r;
  int submatch_id = 0;
  int position = 0;
  tre_pos_and_tags_t *p;
  int *counts = NULL, *offs = NULL;
  int i, add = 0, num_tags = 0, end_tag;
  tre_tnfa_transition_t *transitions, *initial;
  tre_tnfa_t *tnfa = NULL;
  tre_submatch_data_t *submatch_data;
  int *tag_directions = NULL, *tag_min_bounds, *marker_offs = NULL;
  reg_errcode_t errcode;
  tre_mem_t *mem;

  /* Allocate a stack used throughout the compilation process for various
     purposes. */
  stack = stack_new(512, 10240, 128);
  if (stack == NULL)
    return REG_ESPACE;
  mem = tre_mem_new();
  if (mem == NULL)
    {
      stack_destroy(stack);
      return REG_ESPACE;
    }

  DPRINT(("regwncomp: parsing '%*ls'\n", n, regex));
  errcode = parse_re(mem, stack, &tree, regex, n, &submatch_id, &position,
		     cflags);
  if (errcode != REG_OK)
    ERROR_EXIT(errcode);
  preg->re_nsub = submatch_id - 1;

#ifdef REGEX_DEBUG
  ast_print(tree);
#endif /* REGEX_DEBUG */

  /* Allocate the pattern buffer struct to be returned. */

  tnfa = mycalloc(1, sizeof(tre_tnfa_t));
  if (tnfa == NULL)
    ERROR_EXIT(REG_ESPACE);

  /* Set up tags for submatch addressing unless it is disabled. */
  if ((cflags & REG_NOSUB) == 0)
    {
      DPRINT(("regcomp: setting up tags\n"));

      /* Figure out how many tags we will need in the optimized tree. */
      errcode = ast_compute_tag_info(stack, tree);
      if (errcode != REG_OK)
	ERROR_EXIT(errcode);
      num_tags = tree->num_tags;
      DPRINT(("num_tags = %d\n", num_tags));

      if (num_tags > 0)
	{

	  tag_directions = mymalloc(sizeof(*tag_directions) * (num_tags + 1));
	  if (tag_directions == NULL)
	    ERROR_EXIT(REG_ESPACE);
	  tnfa->tag_directions = tag_directions;
	  memset(tag_directions, -1, sizeof(*tag_directions) * (num_tags + 1));

	  marker_offs = mycalloc(num_tags, sizeof(*marker_offs));
	  if (marker_offs == NULL)
	    ERROR_EXIT(REG_ESPACE);
	  tnfa->marker_offs = marker_offs;
	}


      submatch_data = mymalloc(sizeof(*submatch_data) * submatch_id);
      if (submatch_data == NULL)
	ERROR_EXIT(REG_ESPACE);
      tnfa->submatch_data = submatch_data;


      tag_min_bounds = mymalloc(sizeof(*tag_min_bounds) * submatch_id * 2);
      if (tag_min_bounds == NULL)
	ERROR_EXIT(REG_ESPACE);
      tnfa->tag_min_bounds = tag_min_bounds;

      for (i = 0; i < submatch_id * 2; i++)
	tag_min_bounds[i] = -1;

      errcode = ast_add_tags(mem, stack, tree, submatch_id, submatch_data,
			     tag_directions, &end_tag, tag_min_bounds,
			     marker_offs);
      if (errcode != REG_OK)
	ERROR_EXIT(errcode);

#ifdef REGEX_DEBUG
      /* XXX - write a function which describes everything. */
      for (i = 0; i < submatch_id; i++)
	DPRINT(("pmatch[%d] = {t%d - %d, t%d - %d}\n", i,
		submatch_data[i].so_tag, submatch_data[i].so_offset,
		submatch_data[i].eo_tag, submatch_data[i].eo_offset));
      for (i = 0; i < num_tags; i++)
	DPRINT(("t%d is %s\n", i,
		tag_directions[i] == TAG_MINIMIZE ?
		"minimized" : "maximized"));
#endif /* REGEX_DEBUG */
    }

  /* Add a dummy node for the final state.
     XXX - For certain patterns this dummy node can be optimized away,
           for example "a*" or "ab*".   Figure out a simple way to detect
           this possibility. */
#if 1
  tmp_ast_l = tree;
  tmp_ast_r = ast_new_literal(mem, 0, 0, position++);
  if (tmp_ast_r == NULL)
    ERROR_EXIT(REG_ESPACE);

  tree = ast_new_catenation(mem, tmp_ast_l, tmp_ast_r);
  if (tree == NULL)
    ERROR_EXIT(REG_ESPACE);

#endif

#ifdef REGEX_DEBUG
  ast_print(tree);
#endif /* REGEX_DEBUG */

  errcode = ast_compute_nfl(mem, stack, tree);
  if (errcode != REG_OK)
    ERROR_EXIT(errcode);


  counts = mymalloc(sizeof(int) * position);
  if (counts == NULL)
    ERROR_EXIT(REG_ESPACE);


  offs = mymalloc(sizeof(int) * position);
  if (offs == NULL)
    ERROR_EXIT(REG_ESPACE);

  for (i = 0; i < position; i++)
    counts[i] = 0;
  ast_to_efree_tnfa(tree, NULL, counts, NULL);

  add = 0;
  for (i = 0; i < position; i++)
    {
      offs[i] = add;
      add += counts[i] + 1;
      counts[i] = 0;
    }

  transitions = mycalloc(add + 1, sizeof(*transitions));
  if (transitions == NULL)
    ERROR_EXIT(REG_ESPACE);
  tnfa->transitions = transitions;
  tnfa->num_transitions = add;

  DPRINT(("transitions %p\n", transitions));
  errcode = ast_to_efree_tnfa(tree, transitions, counts, offs);
  if (errcode != REG_OK)
    ERROR_EXIT(errcode);

  p = tree->firstpos;
  i = 0;
  DPRINT(("initial:"));
  while (p->position >= 0)
    {
      i++;

#ifdef REGEX_DEBUG
      {
	int *tags;

	DPRINT((" %d", p->position));
	tags = p->tags;
	if (tags != NULL)
	  {
	    if (*tags >= 0)
	      DPRINT(("/"));
	    while (*tags >= 0)
	      {
		DPRINT(("%d", *tags));
		tags++;
		if (*tags >= 0)
		  DPRINT((","));
	      }
	  }
	DPRINT((", assert %d. ", p->assertions));
      }
#endif /* REGEX_DEBUG */

      p++;
    }
  DPRINT(("\n"));


  initial = mycalloc(i + 1, sizeof(tre_tnfa_transition_t));
  if (initial == NULL)
    ERROR_EXIT(REG_ESPACE);
  tnfa->initial = initial;

  i = 0;
  for (p = tree->firstpos; p->position >= 0; p++)
    {
      int j;
      initial[i].state = transitions + offs[p->position];
      initial[i].state_id = p->position;
      if (p->tags != NULL)
	{
	  /* Copy the array p->tags, it's allocated in a tre_mem object. */
	  for (j = 0; p->tags[j] >= 0; j++);

	  initial[i].tags = mymalloc(sizeof(*p->tags) * (j + 1));
	  if (initial[i].tags == NULL)
	    ERROR_EXIT(REG_ESPACE);
	  memcpy(initial[i].tags, p->tags, sizeof(*p->tags) * (j + 1));
	}
      else
	initial[i].tags = NULL;
      initial[i].assertions = p->assertions;
      i++;
    }
  initial[i].state = NULL;

  tnfa->num_transitions = add;
  tnfa->final = transitions + offs[tree->lastpos[0].position];
  tnfa->num_tags = num_tags;
  tnfa->end_tag = end_tag;
  tnfa->num_submatches = submatch_id;
  tnfa->num_states = position;
  tnfa->cflags = cflags;

  DPRINT(("final state %p\n", (void *)tnfa->final));

  tre_mem_free(mem);
  stack_destroy(stack);

  myfree(counts);

  myfree(offs);

  preg->value = tnfa;
  return REG_OK;

 error_exit:
  /* Free everything that was allocated and return the error code. */
  tre_mem_free(mem);
  if (stack != NULL)
    stack_destroy(stack);
  if (counts != NULL) {

    myfree(counts); }
  if (offs != NULL) {

    myfree(offs);}
  preg->value = tnfa;
  regfree(preg);
  return errcode;
}

int
regwcomp(regex_t *preg, const wchar_t *regex, int cflags)
{
  return regwncomp(preg, regex, wcslen(regex), cflags);
}

int
regncomp(regex_t *preg, const char *regex, size_t n, int cflags)
{
  wchar_t *wregex;
  int wlen, ret;


  wregex = mymalloc(sizeof(wchar_t) * (n + 1));
  if (wregex == NULL)
    return REG_ESPACE;

  /* If the current locale uses the standard single byte encoding of
     characters, we don't do a multibyte string conversion.  If we did,
     many applications which use the default locale would break since
     the default "C" locale uses the 7-bit ASCII character set, and
     all characters with the eighth bit set would be considered invalid. */
  if (MB_CUR_MAX == 1)
    {
      int i;
      const unsigned char *str = (unsigned char *)regex;
      wchar_t *wstr = wregex;

      for (i = 0; i < n; i++)
	*(wstr++) = *(str++);
      wlen = n;
    }
  else
    {
      mbstate_t state;
      int consumed;
      wchar_t *wcptr = wregex;
      memset(&state, '\0', sizeof(state));
      while (n > 0)
	{
	  consumed = my_mbrtowc(wcptr, regex, n, &state);

	  switch (consumed)
	    {
	    case 0:
	      if (*regex == '\0')
		consumed = 1;
	      else
		{

		  myfree(wregex);
		  return REG_BADPAT;
		}
	      break;
	    case -1:
	      DPRINT(("mbrtowc: error %d: %s.\n", errno, strerror(errno)));

	      myfree(wregex);
	      return REG_BADPAT;
	      break;
	    case -2:
	      /* The last character wasn't complete.  Let's not call it a
		 fatal error. */
	      consumed = n;
	      break;
	    }
	  regex += consumed;
	  n -= consumed;
	  wcptr++;
	}
      wlen = wcptr - wregex;
    }

  wregex[wlen] = L'\0';
  ret = regwncomp(preg, wregex, wlen, cflags);

  myfree(wregex);

  return ret;
}

int
regcomp(regex_t *preg, const char *regex, int cflags)
{
  return regncomp(preg, regex, strlen(regex), cflags);
}


/* Fills the POSIX.2 regmatch_t array according to the TNFA tag and match
   endpoint values. */
static void
fill_pmatch(size_t nmatch, regmatch_t pmatch[],
	    const tre_tnfa_t *tnfa, int *tags, int match_eo)
{
  tre_submatch_data_t *submatch_data;
  int i, marker_tag, marker_boundary, abs_marker_tag, marker_set;

  i = 0;
  if (match_eo >= 0)
    {
      /* Construct submatch offsets from the tags. */
      DPRINT(("end tag = t%d = %d\n", tnfa->end_tag, match_eo));
      submatch_data = tnfa->submatch_data;
      while (i < tnfa->num_submatches && i < nmatch)
	{
	  if (submatch_data[i].so_tag == tnfa->end_tag)
	    pmatch[i].rm_so = match_eo - submatch_data[i].so_offset;
	  else
	    pmatch[i].rm_so =
	      tags[submatch_data[i].so_tag] - submatch_data[i].so_offset;

	  if (submatch_data[i].eo_tag == tnfa->end_tag)
	    pmatch[i].rm_eo = match_eo - submatch_data[i].eo_offset;
	  else
	    pmatch[i].rm_eo =
	      tags[submatch_data[i].eo_tag] - submatch_data[i].eo_offset;

	  if (tnfa->tag_min_bounds[i * 2] >= 0
	      && pmatch[i].rm_so < tags[tnfa->tag_min_bounds[i * 2]])
	    {
	      DPRINT(("Submatch %d out of bounds (%d < %d), resetting\n",
		      i, pmatch[i].rm_so,
		      tags[tnfa->tag_min_bounds[i * 2]]));
	      pmatch[i].rm_so = -1;
	      pmatch[i].rm_eo = -1;
	    }

	  if (submatch_data[i].marker_tag != 0)
	    {
	      marker_boundary = submatch_data[i].marker_boundary;
	      marker_tag = submatch_data[i].marker_tag;
	      abs_marker_tag = marker_tag > 0 ? marker_tag : -marker_tag;
	      DPRINT(("marker_boundary = t%d = %d, marker_tag = t%d = %d, "
		      "offs = %d\n",
		      marker_boundary, tags[marker_boundary],
		      marker_tag, tags[abs_marker_tag],
		      tnfa->marker_offs[abs_marker_tag]));

	      /* We consider the marker to be set for this submatch if
		 - the marker tag is used, and
		   - there is no marker boundary tag, or
		   - the marker boundary tag is smaller than or equal to
		     marker tag minus the marker tag offset. */
	      marker_set =
		tags[abs_marker_tag] >= 0
		&& (marker_boundary <= 0
		    || (tags[abs_marker_tag] -
			tnfa->marker_offs[abs_marker_tag]
			>= tags[marker_boundary]));
	      if (marker_tag > 0 && !marker_set)
		{
		  DPRINT(("Marker %d for submatch %d is false, resetting.\n",
			  marker_tag, i / 2));
		  pmatch[i].rm_so = -1;
		  pmatch[i].rm_eo = -1;
		}
	    }

	  DPRINT(("pmatch[%d] = {t%d - %d = %d, t%d - %d = %d}\n", i,
		  submatch_data[i].so_tag, submatch_data[i].so_offset,
		  pmatch[i].rm_so,
		  submatch_data[i].eo_tag, submatch_data[i].eo_offset,
		  pmatch[i].rm_eo));

	  i++;
	}
    }
  while (i < nmatch)
    {
      pmatch[i].rm_so = -1;
      pmatch[i].rm_eo = -1;
      i++;
    }
}

int tnfa_match(const tre_tnfa_t *tnfa, const void *string, size_t len,
	       tre_str_type_t type, size_t nmatch, regmatch_t pmatch[],
	       int eflags)
{
  reg_errcode_t status;
  int *tags = NULL, eo;
  if (nmatch == 0)
    eflags |= REG_NOTAGS;
  if (tnfa->num_tags > 0 && !(eflags & REG_NOTAGS))
    {

      tags = mymalloc(sizeof(*tags) * tnfa->num_tags);
      if (tags == NULL)
	return REG_ESPACE;
    }
  status = tnfa_execute(tnfa, string, len, type, tags, eflags, &eo);
  if (status == REG_OK)
    fill_pmatch(nmatch, pmatch, tnfa, tags, eo);
  if (tnfa->num_tags > 0 && !(eflags & REG_NOTAGS))

    myfree(tags);
  return status;
}

int
regnexec(const regex_t *preg, const char *str, size_t len,
	 size_t nmatch, regmatch_t pmatch[], int eflags)
{
  if (MB_CUR_MAX == 1)
    return tnfa_match(preg->value, str, len, STR_BYTE, nmatch, pmatch, eflags);
  else
    return tnfa_match(preg->value, str, len, STR_MBS, nmatch, pmatch, eflags);
}

int
regexec(const regex_t *preg, const char *str,
	size_t nmatch, regmatch_t pmatch[], int eflags)
{
  return regnexec(preg, str, -1, nmatch, pmatch, eflags);
}


int
regwnexec(const regex_t *preg, const wchar_t *str, size_t len,
	  size_t nmatch, regmatch_t pmatch[], int eflags)
{
  return tnfa_match(preg->value, str, len, STR_WIDE, nmatch, pmatch, eflags);
}

int
regwexec(const regex_t *preg, const wchar_t *str,
	 size_t nmatch, regmatch_t pmatch[], int eflags)
{
  return regwnexec(preg, str, -1, nmatch, pmatch, eflags);
}


/* Error message strings for error codes listed in `regex.h'.  This list
   needs to be in sync with the codes listed there, naturally. */
static const wchar_t *error_messages[] =
  { L"No error",                      /* REG_OK */
    L"No match",                      /* REG_NOMATCH */
    L"Invalid regexp",                /* REG_BADPAT */
    L"Unknown collating element",     /* REG_ECOLLATE */
    L"Unknown character class name",  /* REG_ECTYPE */
    L"Trailing backslash",            /* REG_EESCAPE */
    L"XXX",                           /* REG_ESUBREG */
    L"Missing ']'",                   /* REG_EBRACK */
    L"Missing ')'",                   /* REG_EPAREN */
    L"Missing '}'",                   /* REG_EBRACE */
    L"Invalid contents of {}",        /* REG_BADBR */
    L"Invalid character range",       /* REG_ERANGE */
    L"Out of memory",                 /* REG_ESPACE */
    L"XXX"                            /* REG_BADRPT */
  };

size_t
regwerror(int errcode, const regex_t *preg, wchar_t *errbuf,
	  size_t errbuf_size)
{
  const wchar_t *err;
  size_t err_len;

  if (errcode >= 0
      && errcode < (sizeof(error_messages) / sizeof(*error_messages)))
    err = error_messages[errcode];
  else
    err = L"Unknown error";

  err_len = wcslen(err) + 1;
  if (errbuf_size > 0 && errbuf != NULL)
    {
      if (err_len > errbuf_size)
	{
	  wcsncpy(errbuf, err, errbuf_size - 1);
	  errbuf[errbuf_size - 1] = L'\0';
	}
      else
	{
	  wcscpy(errbuf, err);
	}
    }
  return err_len;
}

size_t
regerror(int errcode, const regex_t *preg, char *errbuf,
	 size_t errbuf_size)
{
  wchar_t *err = NULL;
  size_t err_len;
  int get_str = errbuf_size > 0 && errbuf != NULL;

  if (get_str) {

    err = mymalloc(sizeof(*err) * errbuf_size); }
  err_len = regwerror(errcode, preg, err, errbuf_size);
  if (get_str)
    {
      const wchar_t *errptr = err;
#if defined HAVE_WCSRTOMBS
      mbstate_t state;
      memset(&state, '\0', sizeof(state));
      wcsrtombs(errbuf, &errptr, errbuf_size, &state);
#elif defined HAVE_WCSTOMBS
      wcstombs(errbuf, errptr, errbuf_size);
#endif
      errbuf[errbuf_size - 1] = '\0';
    }

  myfree(err);

  return err_len;
}


void
regfree(regex_t *preg)
{
  tre_tnfa_t *tnfa;
  int i;
  tre_tnfa_transition_t *trans;

  tnfa = preg->value;
  if (tnfa == NULL)
    return;

  for (i = 0; i < tnfa->num_transitions; i++)
    if (tnfa->transitions[i].state != NULL)
      {

	myfree(tnfa->transitions[i].tags);

	myfree(tnfa->transitions[i].neg_classes);
      }

  myfree(tnfa->transitions);

  if (tnfa->initial != NULL)
    {
      for (trans = tnfa->initial; trans->state != NULL; trans++) {

	myfree(trans->tags); }

      myfree(tnfa->initial);
    }

  myfree(tnfa->submatch_data);

  myfree(tnfa->tag_directions);

  myfree(tnfa->marker_offs);

  myfree(tnfa->tag_min_bounds);

  myfree(tnfa);
}

/* EOF */





